
CHAR_MAX = 127

CHAR_MIN = -128

DBL_MAX = 1.7976931348623157e+308

DBL_MIN = 2.2250738585072014e-308

FLT_MAX = 3.4028234663852886e+38

FLT_MIN = 1.1754943508222875e-38

INT_MAX = 2147483647

INT_MIN = -2147483648

LLONG_MAX = 9223372036854775807

LLONG_MIN = -9223372036854775808

LONG_MAX = 2147483647

LONG_MIN = -2147483648

PY_SSIZE_T_MAX = 2147483647

PY_SSIZE_T_MIN = -2147483648

SHRT_MAX = 32767

SHRT_MIN = -32768

SIZEOF_PYGC_HEAD = 16

UCHAR_MAX = 255

UINT_MAX = 4294967295

ULLONG_MAX = 18446744073709551615

ULONG_MAX = 4294967295

USHRT_MAX = 65535

__loader__ = "<_frozen_importlib.ExtensionFileLoader object at 0x00C98DD0>"

def _pending_threadfunc(*args,**kw):
    pass

class _test_structmembersType(object):
    pass

def _test_thread_state(*args,**kw):
    pass

def argparsing(*args,**kw):
    pass

def code_newempty(*args,**kw):
    pass

def codec_incrementaldecoder(*args,**kw):
    pass

def codec_incrementalencoder(*args,**kw):
    pass

def crash_no_current_thread(*args,**kw):
    pass

class error(Exception):
    pass

def exception_print(*args,**kw):
    pass

def getargs_B(*args,**kw):
    pass

def getargs_H(*args,**kw):
    pass

def getargs_I(*args,**kw):
    pass

def getargs_K(*args,**kw):
    pass

def getargs_L(*args,**kw):
    pass

def getargs_Z(*args,**kw):
    pass

def getargs_Z_hash(*args,**kw):
    pass

def getargs_b(*args,**kw):
    pass

def getargs_c(*args,**kw):
    pass

def getargs_h(*args,**kw):
    pass

def getargs_i(*args,**kw):
    pass

def getargs_k(*args,**kw):
    pass

def getargs_keyword_only(*args,**kw):
    pass

def getargs_keywords(*args,**kw):
    pass

def getargs_l(*args,**kw):
    pass

def getargs_n(*args,**kw):
    pass

def getargs_p(*args,**kw):
    pass

def getargs_s(*args,**kw):
    pass

def getargs_s_hash(*args,**kw):
    pass

def getargs_s_star(*args,**kw):
    pass

def getargs_tuple(*args,**kw):
    pass

def getargs_u(*args,**kw):
    pass

def getargs_u_hash(*args,**kw):
    pass

def getargs_w_star(*args,**kw):
    pass

def getargs_y(*args,**kw):
    pass

def getargs_y_hash(*args,**kw):
    pass

def getargs_y_star(*args,**kw):
    pass

def getargs_z(*args,**kw):
    pass

def getargs_z_hash(*args,**kw):
    pass

def getargs_z_star(*args,**kw):
    pass

class instancemethod(object):
    pass

def make_exception_with_doc(*args,**kw):
    pass

def make_memoryview_from_NULL_pointer(*args,**kw):
    pass

def parse_tuple_and_keywords(*args,**kw):
    pass

def pytime_object_to_time_t(*args,**kw):
    pass

def pytime_object_to_timespec(*args,**kw):
    pass

def pytime_object_to_timeval(*args,**kw):
    pass

def raise_exception(*args,**kw):
    pass

def raise_memoryerror(*args,**kw):
    pass

def run_in_subinterp(*args,**kw):
    pass

def set_exc_info(*args,**kw):
    pass

def test_L_code(*args,**kw):
    pass

def test_Z_code(*args,**kw):
    pass

def test_capsule(*args,**kw):
    pass

def test_config(*args,**kw):
    pass

def test_datetime_capi(*args,**kw):
    pass

def test_dict_iteration(*args,**kw):
    pass

def test_empty_argparse(*args,**kw):
    pass

def test_k_code(*args,**kw):
    pass

def test_lazy_hash_inheritance(*args,**kw):
    pass

def test_list_api(*args,**kw):
    pass

def test_long_and_overflow(*args,**kw):
    pass

def test_long_api(*args,**kw):
    pass

def test_long_as_double(*args,**kw):
    pass

def test_long_as_size_t(*args,**kw):
    pass

def test_long_long_and_overflow(*args,**kw):
    pass

def test_long_numbits(*args,**kw):
    pass

def test_longlong_api(*args,**kw):
    pass

def test_null_strings(*args,**kw):
    pass

def test_s_code(*args,**kw):
    pass

def test_string_from_format(*args,**kw):
    pass

def test_string_to_double(*args,**kw):
    pass

def test_u_code(*args,**kw):
    pass

def test_unicode_compare_with_ascii(*args,**kw):
    pass

def test_widechar(*args,**kw):
    pass

def test_with_docstring(*args,**kw):
    """This is a pretty normal docstring."""
    pass

def traceback_print(*args,**kw):
    pass

def unicode_aswidechar(*args,**kw):
    pass

def unicode_aswidecharstring(*args,**kw):
    pass

def unicode_encodedecimal(*args,**kw):
    pass

def unicode_transformdecimaltoascii(*args,**kw):
    pass



"""(Extremely) low-level import machinery bits as used by importlib and imp."""


class __loader__(object):pass

def _fix_co_filename(*args,**kw):
    raise NotImplementedError("%s:not implemented" % ('_imp.py:_fix_co_filename'))

def acquire_lock(*args,**kw):
    """acquire_lock() -> None    Acquires the interpreter's import lock for the current thread.
    This lock should be used by import hooks to ensure thread-safety
    when importing modules.
    On platforms without threads, this function does nothing."""
    pass   #assume we are a platform without threads
    #raise NotImplementedError("%s:not implemented" % ('_imp.py:acquire_lock'))

def extension_suffixes(*args,**kw):
    """extension_suffixes() -> list of strings    Returns the list of file suffixes used to identify extension modules."""
    return ['.pyd']

def get_frozen_object(*args,**kw):
    raise NotImplementedError("%s:not implemented" % ('_imp.py:get_frozen_object'))

def init_builtin(module,*args,**kw):
    return __import__(module)

def init_frozen(*args,**kw):
    raise NotImplementedError("%s:not implemented" % ('_imp.py:init_frozen'))

def is_builtin(*args,**kw):
    raise NotImplementedError("%s:not implemented" % ('_imp.py:is_builtin'))

def is_frozen(*args,**kw):
    raise NotImplementedError("%s:not implemented" % ('_imp.py:is_frozen'))

def is_frozen_package(*args,**kw):
    raise NotImplementedError("%s:not implemented" % ('_imp.py:is_frozen_package'))

def load_dynamic(*args,**kw):
    raise NotImplementedError("%s:not implemented" % ('_imp.py:load_dynamic'))

def lock_held(*args,**kw):
    """lock_held() -> boolean    Return True if the import lock is currently held, else False.
    On platforms without threads, return False."""

    return False
    #raise NotImplementedError("%s:not implemented" % ('_imp.py:lock_held'))

def release_lock(*args,**kw):
    """release_lock() -> None    Release the interpreter's import lock.
    On platforms without threads, this function does nothing."""

    pass  #assume no threads
    #raise NotImplementedError("%s:not implemented" % ('_imp.py:release_lock'))



from browser import window, alert

def _randint(a, b):
    return int(window.Math.random()*(b-a+1)+a)
    
def _rand_with_seed(x, rand_obj):
    # if rand_obj.state is not a safe integer, Math.sin will return the same
    # result for consecutive values : use the rest of division by 360
    degrees = rand_obj._state % 360
    x = window.Math.sin(degrees/(2*window.Math.PI)) * 10000
    # Adding 1 is not reliable because of current integer implementation
    # If rand_obj._state is not a "safe integer" in the range [-2**53, 2**53]
    # the increment between 2 different values is a power of 2
    # It is stored in an attribute of rand_obj to avoid having to compute it
    # for each iteration
    if not hasattr(rand_obj, 'incr'):
        rand_obj.incr = 1
    rand_obj._state += rand_obj.incr
    return x - window.Math.floor(x)

def _urandom(n, rand_obj=None):
    """urandom(n) -> str    
    Return n random bytes suitable for cryptographic use."""
    
    if rand_obj is None or rand_obj._state is None:
        randbytes= [_randint(0,255) for i in range(n)]
    else:
        randbytes= []
        for i in range(n):
            randbytes.append(int(256*_rand_with_seed(i, rand_obj)))
    return bytes(randbytes)
    
class Random:
    """Random number generator base class used by bound module functions.

    Used to instantiate instances of Random to get generators that don't
    share state.

    Class Random can also be subclassed if you want to use a different basic
    generator of your own devising: in that case, override the following
    methods:  random(), seed(), getstate(), and setstate().
    Optionally, implement a getrandbits() method so that randrange()
    can cover arbitrarily large ranges.

    """
    #random
    #seed
    #getstate
    #setstate


    VERSION = 3     # used by getstate/setstate

    def __init__(self, x=None):
        """Initialize an instance.

        Optional argument x controls seeding, as for Random.seed().
        """

        self._state=x

    def seed(self, a=None, version=2):
        """Initialize internal state from hashable object.

        None or no argument seeds from current time or from an operating
        system specific randomness source if available.

        For version 2 (the default), all of the bits are used if *a* is a str,
        bytes, or bytearray.  For version 1, the hash() of *a* is used instead.

        If *a* is an int, all bits are used.

        """

        self._state=a
        self.gauss_next = None

    def getstate(self):
        """Return internal state; can be passed to setstate() later."""
        return self._state

    def setstate(self, state):
        """Restore internal state from object returned by getstate()."""
        self._state=state

    def random(self):
        """Get the next random number in the range [0.0, 1.0)."""
        return window.Math.random()

    def getrandbits(self, k):
        """getrandbits(k) -> x.  Generates a long int with k random bits."""
        if k <= 0:
            raise ValueError('number of bits must be greater than zero')
        if k != int(k):
            raise TypeError('number of bits should be an integer')
        numbytes = (k + 7) // 8                       # bits / 8 and rounded up
        x = int.from_bytes(_urandom(numbytes, self), 'big')
            
        return x >> (numbytes * 8 - k)                # trim excess bits



from browser import document, window


def redirect(ev):
    """Create the links"""
    document['navbar_link'].value = ev.target.id
    document['form_base'].action = 'submit'
    document['form_base'].submit()


def mark_as_active(pathname):
    """Mark the active page in the navbar"""
    for id_page in ('home', 'data', 'credits', 'contact', 'info'):
        if id_page in pathname:
            document[id_page].class_name = 'active'
        else:
            document[id_page].class_name = ''
    if 'show' in pathname:
        document['home'].class_name = 'active'


# This highlight the appropriate link in the navbar
mark_as_active(window.location.pathname)


# Events binding
document['home_main'].bind('click', redirect)
document['home'].bind('click', redirect)
document['data'].bind('click', redirect)
document['credits'].bind('click', redirect)
document['contact'].bind('click', redirect)
document['info'].bind('click', redirect)



""" Minimalist version for Brython. Not much can be known about the platform
with window.navigator.
"""

from browser import window

def architecture(*args, **kw):
    return "<unknown>", window.navigator.platform

def machine(*args, **kw):
    return ''

def node(*args, **kw):
    return ''

def platform(*args, **kw):
    return window.navigator.platform

def processor(*args, **kw):
    return ''

def python_build():
    return ('.'.join(map(str, __BRYTHON__.implementation[:-1])), 
        __BRYTHON__.compiled_date)

def python_compiler():
    return ''

def python_branch():
    return ''

def python_implementation():
    return 'Brython'

def python_revision():
    return ''

def python_version():
    return '.'.join(map(str, __BRYTHON__.version_info[:3]))

def python_version_tuple():
    return __BRYTHON__.version_info[:3]

def release():
    return ''

def system():
    return window.navigator.platform

def system_alias(*args, **kw):
    return window.navigator.platform

def uname():
    from collections import namedtuple
    klass = namedtuple('uname_result', 
        'system node release version machine processor')
    return klass(window.navigator.platform, '', '', '', '', '')


"""string helper module"""

import re

class __loader__(object):
    pass

def formatter_field_name_split(fieldname):
    """split the argument as a field name"""
    _list=[]
    for _name in fieldname:
        _parts = _name.split('.')
        for _item in _parts:
            is_attr=False  #fix me
            if re.match('\d+', _item):
               _list.append((int(_item), is_attr))
            else:
               _list.append((_item, is_attr))

    return _list[0][0], iter(_list[1:])

def formatter_parser(*args,**kw):
    """parse the argument as a format string"""

    assert len(args)==1
    assert isinstance(args[0], str)

    _result=[]
    for _match in re.finditer("([^{]*)?(\{[^}]*\})?", args[0]):
        _pre, _fmt = _match.groups()
        if _fmt is None:
           _result.append((_pre, None, None, None))
        elif _fmt == '{}':
           _result.append((_pre, '', '', None))
        else:
           _m=re.match("\{([^!]*)!?(.*)?\}", _fmt)
           _name=_m.groups(0)
           _flags=_m.groups(1)

           _result.append((_pre, _name, _flags, None))

    return _result



import random
from browser import document as doc, html

def getMousePosition(e):
    if e is None:
       e=win.event

    if e.pageX or e.pageY:
       return {'x': e.pageX, 'y': e.pageY}

    if e.clientX or e.clientY:
       _posx=e.clientX + doc.body.scrollLeft + doc.documentElement.scrollLeft;
       _posy=e.clientY + doc.body.scrollTop + doc.documentElement.scrollTop;
       return {'x': _posx, 'y': _posy}
      
    return {'x': 0, 'y': 0}

class Widget:
  def __init__(self, element, type, id=None):
      self._element=element

      if id is None:
         self._element.id='%s_%s' % (type, int(100000*random.random()))
      else:
         self._element.id=id

  def get_id(self):
      return self._element.id

  def attach(self, element_id):
      """ append this DOM component to DOM element element_id"""
      #document[element_id] <= self._element   #this doesn't work :(
      #doc is actually the global 'doc' not the one we imported from browser :(
      doc[element_id] <= self._element

  def show(self):
      self._element.display='block'

  def hide(self):
      self._element.display='none'

class DraggableWidget(Widget):
  def __init__(self, element, type, id=None):
      Widget.__init__(self, element, type, id)

      def drag(e):
          self._element.style.top='%spx' % (e.clientY - self._deltaY)
          self._element.style.left='%spx' % (e.clientX - self._deltaX)

      def mouseDown(e):
          self._element.style.position='absolute'
          self._deltaX=e.clientX - self._element.offsetLeft
          self._deltaY=e.clientY - self._element.offsetTop
          doc.bind('mousemove', drag)

      def mouseUp(e):
          doc.unbind('mousemove')

      self._element.bind('mousedown', mouseDown)
      self._element.bind('mouseup', mouseUp)



"""This module provides access to the garbage collector for reference cycles.

enable() -- Enable automatic garbage collection.
disable() -- Disable automatic garbage collection.
isenabled() -- Returns true if automatic collection is enabled.
collect() -- Do a full collection right now.
get_count() -- Return the current collection counts.
set_debug() -- Set debugging flags.
get_debug() -- Get debugging flags.
set_threshold() -- Set the collection thresholds.
get_threshold() -- Return the current the collection thresholds.
get_objects() -- Return a list of all objects tracked by the collector.
is_tracked() -- Returns true if a given object is tracked.
get_referrers() -- Return the list of objects that refer to an object.
get_referents() -- Return the list of objects that an object refers to.
"""


DEBUG_COLLECTABLE = 2

DEBUG_LEAK = 38

DEBUG_SAVEALL = 32

DEBUG_STATS = 1

DEBUG_UNCOLLECTABLE = 4

class __loader__:
    pass

callbacks = []

def collect(*args,**kw):
    """collect([generation]) -> n    
    With no arguments, run a full collection.  The optional argument
    may be an integer specifying which generation to collect.  A ValueError
    is raised if the generation number is invalid.
    
    The number of unreachable objects is returned.
    """
    pass

def disable(*args,**kw):
    """disable() -> None    
    Disable automatic garbage collection.
    """
    pass

def enable(*args,**kw):
    """enable() -> None    
    Enable automatic garbage collection.
    """
    pass

garbage = []

def get_count(*args,**kw):
    """get_count() -> (count0, count1, count2)    
    Return the current collection counts
    """
    pass

def get_debug(*args,**kw):
    """get_debug() -> flags    
    Get the garbage collection debugging flags.
    """
    pass

def get_objects(*args,**kw):
    """get_objects() -> [...]    
    Return a list of objects tracked by the collector (excluding the list
    returned).
    """
    pass

def get_referents(*args,**kw):
    """get_referents(*objs) -> list    Return the list of objects that are directly referred to by objs."""
    pass

def get_referrers(*args,**kw):
    """get_referrers(*objs) -> list    Return the list of objects that directly refer to any of objs."""
    pass

def get_threshold(*args,**kw):
    """get_threshold() -> (threshold0, threshold1, threshold2)    
    Return the current collection thresholds
    """
    pass

def is_tracked(*args,**kw):
    """is_tracked(obj) -> bool    
    Returns true if the object is tracked by the garbage collector.
    Simple atomic objects will return false.
    """
    pass

def isenabled(*args,**kw):
    """isenabled() -> status    
    Returns true if automatic garbage collection is enabled.
    """
    pass

def set_debug(*args,**kw):
    """set_debug(flags) -> None    
    Set the garbage collection debugging flags. Debugging information is
    written to sys.stderr.
    
    flags is an integer and can have the following bits turned on:
    
      DEBUG_STATS - Print statistics during collection.
      DEBUG_COLLECTABLE - Print collectable objects found.
      DEBUG_UNCOLLECTABLE - Print unreachable but uncollectable objects found.
      DEBUG_SAVEALL - Save objects to gc.garbage rather than freeing them.
      DEBUG_LEAK - Debug leaking programs (everything but STATS).
    """
    pass

def set_threshold(*args,**kw):
    """set_threshold(threshold0, [threshold1, threshold2]) -> None    
    Sets the collection thresholds.  Setting threshold0 to zero disables
    collection.
    """
    pass



# session storage in browser
import sys
from javascript import JSObject
from browser import window
from .local_storage import LocalStorage

has_session_storage = hasattr(window, 'sessionStorage')

class SessionStorage(LocalStorage):

    storage_type = "session_storage"

    def __init__(self):
        if not has_session_storage:
            raise EnvironmentError("SessionStorage not available")
        self.store = window.sessionStorage

if has_session_storage:
    storage = SessionStorage()



from . import widget
from browser import doc,html

class Slider(widget.Widget):

  def __init__(self, id=None, label=False):

      self._div_shell=html.DIV(Class="ui-slider ui-slider-horizontal ui-widget ui-widget-content ui-corner-all")

      widget.Widget.__init__(self, self._div_shell, 'slider', id)

      self._handle=html.A(Class="ui-slider-handle ui-state-default ui-corner-all",
                          Href='#', style={'left': '0px'})
      self._value=0
      self._isMouseDown=False

      def startSlide(e):
          self._isMouseDown=True
          self._upperBound = self._div_shell.offsetWidth - self._handle.offsetWidth

          pos = widget.getMousePosition(e)
          self._startMouseX=pos['x']

          self._lastElementLeft = parseInt(self._handle.style.left)
          updatePosition(e)

      def updatePosition(e):
          pos = widget.getMousePosition(e)
          #print('mose pos',pos)
          _newPos = self._lastElementLeft + pos['x'] - self._startMouseX
          
          _newPos = max(0, _newPos)
          _newPos = min(_newPos, self._upperBound)

          self._handle.style.left = '%spx' % _newPos
          #print('new position',self._handle.style.left)
          self._lastElementLeft = _newPos

      def moving(e):
          if self._isMouseDown:
             updatePosition(e)

      def dropCallback(e):
          self._isMouseDown=False
          self._handle.unbind('mousemove', moving)


      self._handle.bind('mousemove', moving)
      self._handle.bind('mouseup', dropCallback)
      #self._handle.bind('mouseout', dropCallback)
      self._handle.bind('mousedown', startSlide)

      def mouseover(e):
          _class=self._handle.getAttribute('class')
          self._handle.setAttribute('class', '%s %s' % (_class, 'ui-state-hover'))

      def mouseout(e):
          self._isMouseDown=False
          _class=self._handle.getAttribute('class')
          self._handle.setAttribute('class', _class.replace('ui-state-hover', ''))

      self._handle.bind('mouseover', mouseover)
      self._handle.bind('mouseout', mouseout)

      self._div_shell <= self._handle

  def get_value(self):
      return self._value

  #def set_value(self, value):
  #    self._value=value
  #   self._handle.style.left='%spx' % value



from browser import window
import javascript

if hasattr(window, 'WebSocket'):
    supported = True
    WebSocket = javascript.JSConstructor(window.WebSocket)
else:
    supported = False
    def WebSocket(*args,**kw):
        raise NotImplementedError


"""A pure Python implementation of binascii.

Rather slow and buggy in corner cases.
PyPy provides an RPython version too.
"""

# borrowed from https://bitbucket.org/pypy/pypy/src/f2bf94943a41/lib_pypy/binascii.py

class Error(Exception):
    pass

class Done(Exception):
    pass

class Incomplete(Exception):
    pass

def a2b_uu(s):
    if not s:
        return ''
    
    length = (ord(s[0]) - 0x20) % 64

    def quadruplets_gen(s):
        while s:
            try:
                yield ord(s[0]), ord(s[1]), ord(s[2]), ord(s[3])
            except IndexError:
                s += '   '
                yield ord(s[0]), ord(s[1]), ord(s[2]), ord(s[3])
                return
            s = s[4:]

    try:
        result = [''.join(
            [chr((A - 0x20) << 2 | (((B - 0x20) >> 4) & 0x3)),
            chr(((B - 0x20) & 0xf) << 4 | (((C - 0x20) >> 2) & 0xf)),
            chr(((C - 0x20) & 0x3) << 6 | ((D - 0x20) & 0x3f))
            ]) for A, B, C, D in quadruplets_gen(s[1:].rstrip())]
    except ValueError:
        raise Error('Illegal char')
    result = ''.join(result)
    trailingdata = result[length:]
    if trailingdata.strip('\x00'):
        raise Error('Trailing garbage')
    result = result[:length]
    if len(result) < length:
        result += ((length - len(result)) * '\x00')
    return bytes(result, __BRYTHON__.charset)


def b2a_uu(s):
    length = len(s)
    if length > 45:
        raise Error('At most 45 bytes at once')

    def triples_gen(s):
        while s:
            try:
                yield ord(s[0]), ord(s[1]), ord(s[2])
            except IndexError:
                s += '\0\0'
                yield ord(s[0]), ord(s[1]), ord(s[2])
                return
            s = s[3:]

    result = [''.join(
        [chr(0x20 + (( A >> 2                    ) & 0x3F)),
         chr(0x20 + (((A << 4) | ((B >> 4) & 0xF)) & 0x3F)),
         chr(0x20 + (((B << 2) | ((C >> 6) & 0x3)) & 0x3F)),
         chr(0x20 + (( C                         ) & 0x3F))])
              for A, B, C in triples_gen(s)]
    return chr(ord(' ') + (length & 0o77)) + ''.join(result) + '\n'


table_a2b_base64 = {
    'A': 0,
    'B': 1,
    'C': 2,
    'D': 3,
    'E': 4,
    'F': 5,
    'G': 6,
    'H': 7,
    'I': 8,
    'J': 9,
    'K': 10,
    'L': 11,
    'M': 12,
    'N': 13,
    'O': 14,
    'P': 15,
    'Q': 16,
    'R': 17,
    'S': 18,
    'T': 19,
    'U': 20,
    'V': 21,
    'W': 22,
    'X': 23,
    'Y': 24,
    'Z': 25,
    'a': 26,
    'b': 27,
    'c': 28,
    'd': 29,
    'e': 30,
    'f': 31,
    'g': 32,
    'h': 33,
    'i': 34,
    'j': 35,
    'k': 36,
    'l': 37,
    'm': 38,
    'n': 39,
    'o': 40,
    'p': 41,
    'q': 42,
    'r': 43,
    's': 44,
    't': 45,
    'u': 46,
    'v': 47,
    'w': 48,
    'x': 49,
    'y': 50,
    'z': 51,
    '0': 52,
    '1': 53,
    '2': 54,
    '3': 55,
    '4': 56,
    '5': 57,
    '6': 58,
    '7': 59,
    '8': 60,
    '9': 61,
    '+': 62,
    '/': 63,
    '=': 0,
}


def a2b_base64(s):
    if not isinstance(s, (str, bytes)):
        raise TypeError("expected string, got %r" % (s,))
    s = s.rstrip()
    # clean out all invalid characters, this also strips the final '=' padding
    # check for correct padding

    def next_valid_char(s, pos):
        for i in range(pos + 1, len(s)):
            c = s[i]
            if c < 0x7f:
                try:
                    table_a2b_base64[chr(c)]
                    return chr(c)
                except KeyError:
                    pass
        return None
    
    quad_pos = 0
    leftbits = 0
    leftchar = 0
    res = []
    for i, c in enumerate(s):
        if isinstance(c, int):
            c = chr(c)
        if c > '\x7f' or c == '\n' or c == '\r' or c == ' ':
            continue
        if c == '=':
            if quad_pos < 2 or (quad_pos == 2 and next_valid_char(s, i) != '='):
                continue
            else:
                leftbits = 0
                break
        try:
            next_c = table_a2b_base64[c]
        except KeyError:
            continue
        quad_pos = (quad_pos + 1) & 0x03
        leftchar = (leftchar << 6) | next_c
        leftbits += 6
        if leftbits >= 8:
            leftbits -= 8
            res.append((leftchar >> leftbits & 0xff))
            leftchar &= ((1 << leftbits) - 1)
    if leftbits != 0:
        raise Error('Incorrect padding')

    return bytes(''.join([chr(i) for i in res]),__BRYTHON__.charset)
    
table_b2a_base64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"\
    "0123456789+/"

def b2a_base64(s):
    length = len(s)
    final_length = length % 3

    def triples_gen(s):
        while s:
            try:
                yield s[0], s[1], s[2]
            except IndexError:
                s += b'\0\0'
                yield s[0], s[1], s[2]
                return
            s = s[3:]

    a = triples_gen(s[ :length - final_length])

    result = [''.join(
        [table_b2a_base64[( A >> 2                    ) & 0x3F],
         table_b2a_base64[((A << 4) | ((B >> 4) & 0xF)) & 0x3F],
         table_b2a_base64[((B << 2) | ((C >> 6) & 0x3)) & 0x3F],
         table_b2a_base64[( C                         ) & 0x3F]])
              for A, B, C in a]

    final = s[length - final_length:]
    if final_length == 0:
        snippet = ''
    elif final_length == 1:
        a = final[0]
        snippet = table_b2a_base64[(a >> 2 ) & 0x3F] + \
                  table_b2a_base64[(a << 4 ) & 0x3F] + '=='
    else:
        a = final[0]
        b = final[1]
        snippet = table_b2a_base64[(a >> 2) & 0x3F] + \
                  table_b2a_base64[((a << 4) | (b >> 4) & 0xF) & 0x3F] + \
                  table_b2a_base64[(b << 2) & 0x3F] + '='
    return bytes(''.join(result) + snippet + '\n', __BRYTHON__.charset)

def a2b_qp(s, header=False):
    inp = 0
    odata = []
    while inp < len(s):
        if s[inp] == '=':
            inp += 1
            if inp >= len(s):
                break
            # Soft line breaks
            if (s[inp] == '\n') or (s[inp] == '\r'):
                if s[inp] != '\n':
                    while inp < len(s) and s[inp] != '\n':
                        inp += 1
                if inp < len(s):
                    inp += 1
            elif s[inp] == '=':
                # broken case from broken python qp
                odata.append('=')
                inp += 1
            elif s[inp] in hex_numbers and s[inp + 1] in hex_numbers:
                ch = chr(int(s[inp:inp+2], 16))
                inp += 2
                odata.append(ch)
            else:
                odata.append('=')
        elif header and s[inp] == '_':
            odata.append(' ')
            inp += 1
        else:
            odata.append(s[inp])
            inp += 1
    return bytes(''.join(odata), __BRYTHON__.charset)

def b2a_qp(data, quotetabs=False, istext=True, header=False):
    """quotetabs=True means that tab and space characters are always
       quoted.
       istext=False means that \r and \n are treated as regular characters
       header=True encodes space characters with '_' and requires
       real '_' characters to be quoted.
    """
    MAXLINESIZE = 76

    # See if this string is using CRLF line ends
    lf = data.find('\n')
    crlf = lf > 0 and data[lf-1] == '\r'

    inp = 0
    linelen = 0
    odata = []
    while inp < len(data):
        c = data[inp]
        if (c > '~' or
            c == '=' or
            (header and c == '_') or
            (c == '.' and linelen == 0 and (inp+1 == len(data) or
                                            data[inp+1] == '\n' or
                                            data[inp+1] == '\r')) or
            (not istext and (c == '\r' or c == '\n')) or
            ((c == '\t' or c == ' ') and (inp + 1 == len(data))) or
            (c <= ' ' and c != '\r' and c != '\n' and
             (quotetabs or (not quotetabs and (c != '\t' and c != ' '))))):
            linelen += 3
            if linelen >= MAXLINESIZE:
                odata.append('=')
                if crlf: odata.append('\r')
                odata.append('\n')
                linelen = 3
            odata.append('=' + two_hex_digits(ord(c)))
            inp += 1
        else:
            if (istext and
                (c == '\n' or (inp+1 < len(data) and c == '\r' and
                               data[inp+1] == '\n'))):
                linelen = 0
                # Protect against whitespace on end of line
                if (len(odata) > 0 and
                    (odata[-1] == ' ' or odata[-1] == '\t')):
                    ch = ord(odata[-1])
                    odata[-1] = '='
                    odata.append(two_hex_digits(ch))

                if crlf: odata.append('\r')
                odata.append('\n')
                if c == '\r':
                    inp += 2
                else:
                    inp += 1
            else:
                if (inp + 1 < len(data) and
                    data[inp+1] != '\n' and
                    (linelen + 1) >= MAXLINESIZE):
                    odata.append('=')
                    if crlf: odata.append('\r')
                    odata.append('\n')
                    linelen = 0

                linelen += 1
                if header and c == ' ':
                    c = '_'
                odata.append(c)
                inp += 1
    return ''.join(odata)

hex_numbers = '0123456789ABCDEF'
def hex(n):
    if n == 0:
        return '0'
    
    if n < 0:
        n = -n
        sign = '-'
    else:
        sign = ''
    arr = []

    def hex_gen(n):
        """ Yield a nibble at a time. """
        while n:
            yield n % 0x10
            n = n / 0x10

    for nibble in hex_gen(n):
        arr = [hex_numbers[nibble]] + arr
    return sign + ''.join(arr)

def two_hex_digits(n):
    return hex_numbers[n / 0x10] + hex_numbers[n % 0x10]
    

def strhex_to_int(s):
    i = 0
    for c in s:
        i = i * 0x10 + hex_numbers.index(c)
    return i

hqx_encoding = '!"#$%&\'()*+,-012345689@ABCDEFGHIJKLMNPQRSTUVXYZ[`abcdefhijklmpqr'

DONE = 0x7f
SKIP = 0x7e
FAIL = 0x7d
    
table_a2b_hqx = [
    #^@    ^A    ^B    ^C    ^D    ^E    ^F    ^G   
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    #\b    \t    \n    ^K    ^L    \r    ^N    ^O   
    FAIL, FAIL, SKIP, FAIL, FAIL, SKIP, FAIL, FAIL,
    #^P    ^Q    ^R    ^S    ^T    ^U    ^V    ^W   
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    #^X    ^Y    ^Z    ^[    ^\    ^]    ^^    ^_   
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    #      !     "     #     $     %     &     '   
    FAIL, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
    #(     )     *     +     ,     -     .     /   
    0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, FAIL, FAIL,
    #0     1     2     3     4     5     6     7   
    0x0D, 0x0E, 0x0F, 0x10, 0x11, 0x12, 0x13, FAIL,
    #8     9     :     ;     <     =     >     ?   
    0x14, 0x15, DONE, FAIL, FAIL, FAIL, FAIL, FAIL,
    #@     A     B     C     D     E     F     G   
    0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D,
    #H     I     J     K     L     M     N     O   
    0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23, 0x24, FAIL,
    #P     Q     R     S     T     U     V     W   
    0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, FAIL,
    #X     Y     Z     [     \     ]     ^     _   
    0x2C, 0x2D, 0x2E, 0x2F, FAIL, FAIL, FAIL, FAIL,
    #`     a     b     c     d     e     f     g   
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, FAIL,
    #h     i     j     k     l     m     n     o   
    0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, FAIL, FAIL,
    #p     q     r     s     t     u     v     w   
    0x3D, 0x3E, 0x3F, FAIL, FAIL, FAIL, FAIL, FAIL,
    #x     y     z     {     |     }     ~    ^?   
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
    FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL, FAIL,
]

def a2b_hqx(s):
    result = []

    def quadruples_gen(s):
        t = []
        for c in s:
            res = table_a2b_hqx[ord(c)]
            if res == SKIP:
                continue
            elif res == FAIL:
                raise Error('Illegal character')
            elif res == DONE:
                yield t
                raise Done
            else:
                t.append(res)
            if len(t) == 4:
                yield t
                t = []
        yield t
        
    done = 0
    try:
        for snippet in quadruples_gen(s):
            length = len(snippet)
            if length == 4:
                result.append(chr(((snippet[0] & 0x3f) << 2) | (snippet[1] >> 4))) 
                result.append(chr(((snippet[1] & 0x0f) << 4) | (snippet[2] >> 2))) 
                result.append(chr(((snippet[2] & 0x03) << 6) | (snippet[3]))) 
            elif length == 3:
                result.append(chr(((snippet[0] & 0x3f) << 2) | (snippet[1] >> 4))) 
                result.append(chr(((snippet[1] & 0x0f) << 4) | (snippet[2] >> 2))) 
            elif length == 2:
                result.append(chr(((snippet[0] & 0x3f) << 2) | (snippet[1] >> 4))) 
    except Done:
        done = 1
    except Error:
        raise
    return (''.join(result), done)
    # should this return a bytes object?
    #return (bytes(''.join(result), __BRYTHON__.charset), done)

def b2a_hqx(s):
    result =[]

    def triples_gen(s):
        while s:
            try:
                yield ord(s[0]), ord(s[1]), ord(s[2])
            except IndexError:
                yield tuple([ord(c) for c in s])
            s = s[3:]

    for snippet in triples_gen(s):
        length = len(snippet)
        if length == 3:
            result.append(
                hqx_encoding[(snippet[0] & 0xfc) >> 2])
            result.append(hqx_encoding[
                ((snippet[0] & 0x03) << 4) | ((snippet[1] & 0xf0) >> 4)])
            result.append(hqx_encoding[
                (snippet[1] & 0x0f) << 2 | ((snippet[2] & 0xc0) >> 6)])
            result.append(hqx_encoding[snippet[2] & 0x3f])
        elif length == 2:
            result.append(
                hqx_encoding[(snippet[0] & 0xfc) >> 2])
            result.append(hqx_encoding[
                ((snippet[0] & 0x03) << 4) | ((snippet[1] & 0xf0) >> 4)])
            result.append(hqx_encoding[
                (snippet[1] & 0x0f) << 2])
        elif length == 1:
            result.append(
                hqx_encoding[(snippet[0] & 0xfc) >> 2])
            result.append(hqx_encoding[
                ((snippet[0] & 0x03) << 4)])
    return ''.join(result)

crctab_hqx = [
        0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
        0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
        0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
        0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
        0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
        0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
        0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
        0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
        0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
        0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
        0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
        0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
        0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
        0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
        0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
        0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
        0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
        0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
        0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
        0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
        0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
        0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
        0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
        0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
        0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
        0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
        0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
        0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
        0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
        0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
        0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
        0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
]

def crc_hqx(s, crc):
    for c in s:
        crc = ((crc << 8) & 0xff00) ^ crctab_hqx[((crc >> 8) & 0xff) ^ ord(c)]

    return crc

def rlecode_hqx(s):
    """
    Run length encoding for binhex4.
    The CPython implementation does not do run length encoding
    of \x90 characters. This implementation does.
    """
    if not s:
        return ''
    result = []
    prev = s[0]
    count = 1
    # Add a dummy character to get the loop to go one extra round.
    # The dummy must be different from the last character of s.
    # In the same step we remove the first character, which has
    # already been stored in prev.
    if s[-1] == '!':
        s = s[1:] + '?'
    else:
        s = s[1:] + '!'
        
    for c in s:
        if c == prev and count < 255:
            count += 1
        else:
            if count == 1:
                if prev != '\x90':
                    result.append(prev)
                else:
                    result.extend(['\x90', '\x00'])
            elif count < 4:
                if prev != '\x90':
                    result.extend([prev] * count)
                else:
                    result.extend(['\x90', '\x00'] * count)
            else:
                if prev != '\x90':
                    result.extend([prev, '\x90', chr(count)])
                else:
                    result.extend(['\x90', '\x00', '\x90', chr(count)]) 
            count = 1
            prev = c
        
    return ''.join(result)

def rledecode_hqx(s):
    s = s.split('\x90')
    result = [s[0]]
    prev = s[0]
    for snippet in s[1:]:
        count = ord(snippet[0])
        if count > 0:
            result.append(prev[-1] * (count-1))
            prev = snippet
        else:
            result.append('\x90')
            prev = '\x90'
        result.append(snippet[1:])

    return ''.join(result)

crc_32_tab = [
    0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419,
    0x706af48f, 0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4,
    0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07,
    0x90bf1d91, 0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
    0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7, 0x136c9856,
    0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
    0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4,
    0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
    0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3,
    0x45df5c75, 0xdcd60dcf, 0xabd13d59, 0x26d930ac, 0x51de003a,
    0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599,
    0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
    0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190,
    0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f,
    0x9fbfe4a5, 0xe8b8d433, 0x7807c9a2, 0x0f00f934, 0x9609a88e,
    0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
    0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed,
    0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
    0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3,
    0xfbd44c65, 0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
    0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a,
    0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5,
    0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa, 0xbe0b1010,
    0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
    0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17,
    0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6,
    0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615,
    0x73dc1683, 0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
    0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1, 0xf00f9344,
    0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
    0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a,
    0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
    0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1,
    0xa6bc5767, 0x3fb506dd, 0x48b2364b, 0xd80d2bda, 0xaf0a1b4c,
    0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef,
    0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
    0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe,
    0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31,
    0x2cd99e8b, 0x5bdeae1d, 0x9b64c2b0, 0xec63f226, 0x756aa39c,
    0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
    0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b,
    0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
    0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1,
    0x18b74777, 0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
    0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45, 0xa00ae278,
    0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7,
    0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc, 0x40df0b66,
    0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
    0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605,
    0xcdd70693, 0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8,
    0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b,
    0x2d02ef8d
]

def crc32(s, crc=0):
    result = 0
    crc = ~int(crc) & 0xffffffff
    #crc = ~long(crc) & 0xffffffffL
    for c in s:
        crc = crc_32_tab[(crc ^ int(ord(c))) & 0xff] ^ (crc >> 8)
        #crc = crc_32_tab[(crc ^ long(ord(c))) & 0xffL] ^ (crc >> 8)
        #/* Note:  (crc >> 8) MUST zero fill on left

    result = crc ^ 0xffffffff
    
    if result > 2**31:
        result = ((result + 2**31) % 2**32) - 2**31

    return result

def b2a_hex(s):
    result = []
    for char in s:
        c = (ord(char) >> 4) & 0xf
        if c > 9:
            c = c + ord('a') - 10
        else:
            c = c + ord('0')
        result.append(chr(c))
        c = ord(char) & 0xf
        if c > 9:
            c = c + ord('a') - 10
        else:
            c = c + ord('0')
        result.append(chr(c))
    return ''.join(result)

hexlify = b2a_hex

table_hex = [
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    0, 1, 2, 3,  4, 5, 6, 7,  8, 9,-1,-1, -1,-1,-1,-1,
    -1,10,11,12, 13,14,15,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,10,11,12, 13,14,15,-1, -1,-1,-1,-1, -1,-1,-1,-1,
    -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1, -1,-1,-1,-1
]


def a2b_hex(t):
    result = []

    def pairs_gen(s):
        while s:
            try:
                yield table_hex[ord(s[0])], table_hex[ord(s[1])]
            except IndexError:
                if len(s):
                    raise TypeError('Odd-length string')
                return
            s = s[2:]

    for a, b in pairs_gen(t):
        if a < 0 or b < 0:
            raise TypeError('Non-hexadecimal digit found')
        result.append(chr((a << 4) + b))
    return bytes(''.join(result), __BRYTHON__.charset)
    

unhexlify = a2b_hex



"""A collection of string constants.

Public module variables:

whitespace -- a string containing all ASCII whitespace
ascii_lowercase -- a string containing all ASCII lowercase letters
ascii_uppercase -- a string containing all ASCII uppercase letters
ascii_letters -- a string containing all ASCII letters
digits -- a string containing all ASCII decimal digits
hexdigits -- a string containing all ASCII hexadecimal digits
octdigits -- a string containing all ASCII octal digits
punctuation -- a string containing all ASCII punctuation characters
printable -- a string containing all ASCII characters considered printable

"""

import _string

# Some strings for ctype-style character classification
whitespace = ' \t\n\r\v\f'
ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
ascii_uppercase = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
ascii_letters = ascii_lowercase + ascii_uppercase
digits = '0123456789'
hexdigits = digits + 'abcdef' + 'ABCDEF'
octdigits = '01234567'
punctuation = """!"#$%&'()*+,-./:;<=>?@[\]^_`{|}~"""
printable = digits + ascii_letters + punctuation + whitespace

# Functions which aren't available as string methods.

# Capitalize the words in a string, e.g. " aBc  dEf " -> "Abc Def".
def capwords(s, sep=None):
    """capwords(s [,sep]) -> string

    Split the argument into words using split, capitalize each
    word using capitalize, and join the capitalized words using
    join.  If the optional second argument sep is absent or None,
    runs of whitespace characters are replaced by a single space
    and leading and trailing whitespace are removed, otherwise
    sep is used to split and join the words.

    """
    return (sep or ' ').join(x.capitalize() for x in s.split(sep))


####################################################################
import re as _re
from collections import ChainMap

class _TemplateMetaclass(type):
    pattern = r"""
    %(delim)s(?:
      (?P<escaped>%(delim)s) |   # Escape sequence of two delimiters
      (?P<named>%(id)s)      |   # delimiter and a Python identifier
      {(?P<braced>%(id)s)}   |   # delimiter and a braced identifier
      (?P<invalid>)              # Other ill-formed delimiter exprs
    )
    """

    def __init__(cls, name, bases, dct):
        super(_TemplateMetaclass, cls).__init__(name, bases, dct)
        if 'pattern' in dct:
            pattern = cls.pattern
        else:
            pattern = _TemplateMetaclass.pattern % {
                'delim' : _re.escape(cls.delimiter),
                'id'    : cls.idpattern,
                }
        cls.pattern = _re.compile(pattern, cls.flags | _re.VERBOSE)


class Template(metaclass=_TemplateMetaclass):
    """A string class for supporting $-substitutions."""

    delimiter = '$'
    idpattern = r'[_a-z][_a-z0-9]*'
    flags = _re.IGNORECASE

    def __init__(self, template):
        self.template = template

    # Search for $$, $identifier, ${identifier}, and any bare $'s

    def _invalid(self, mo):
        i = mo.start('invalid')
        lines = self.template[:i].splitlines(keepends=True)
        if not lines:
            colno = 1
            lineno = 1
        else:
            colno = i - len(''.join(lines[:-1]))
            lineno = len(lines)
        raise ValueError('Invalid placeholder in string: line %d, col %d' %
                         (lineno, colno))

    def substitute(self, *args, **kws):
        if len(args) > 1:
            raise TypeError('Too many positional arguments')
        if not args:
            mapping = kws
        elif kws:
            mapping = ChainMap(kws, args[0])
        else:
            mapping = args[0]
        # Helper function for .sub()
        def convert(mo):
            # Check the most common path first.
            named = mo.group('named') or mo.group('braced')
            if named is not None:
                val = mapping[named]
                # We use this idiom instead of str() because the latter will
                # fail if val is a Unicode containing non-ASCII characters.
                return '%s' % (val,)
            if mo.group('escaped') is not None:
                return self.delimiter
            if mo.group('invalid') is not None:
                self._invalid(mo)
            raise ValueError('Unrecognized named group in pattern',
                             self.pattern)
        return self.pattern.sub(convert, self.template)

    def safe_substitute(self, *args, **kws):
        if len(args) > 1:
            raise TypeError('Too many positional arguments')
        if not args:
            mapping = kws
        elif kws:
            mapping = ChainMap(kws, args[0])
        else:
            mapping = args[0]
        # Helper function for .sub()
        def convert(mo):
            named = mo.group('named') or mo.group('braced')
            if named is not None:
                try:
                    # We use this idiom instead of str() because the latter
                    # will fail if val is a Unicode containing non-ASCII
                    return '%s' % (mapping[named],)
                except KeyError:
                    return mo.group()
            if mo.group('escaped') is not None:
                return self.delimiter
            if mo.group('invalid') is not None:
                return mo.group()
            raise ValueError('Unrecognized named group in pattern',
                             self.pattern)
        return self.pattern.sub(convert, self.template)



########################################################################
# the Formatter class
# see PEP 3101 for details and purpose of this class

# The hard parts are reused from the C implementation.  They're exposed as "_"
# prefixed methods of str.

# The overall parser is implemented in _string.formatter_parser.
# The field name parser is implemented in _string.formatter_field_name_split

class Formatter:
    def format(self, format_string, *args, **kwargs):
        return format_string.format(*args, **kwargs)



from . import widget
from browser import html

class ProgressBar(widget.Widget):
  def __init__(self, id=None, label=False):
      self._div_shell=html.DIV(Class="ui-progressbar ui-widget ui-widget-content ui-corner-all")
      widget.Widget.__init__(self, self._div_shell, 'progressbar', id)

      self._show_label=label
      if label:
         self._label=html.DIV(Class='progress-label')
         self._div_shell <= self._label

      self._bar=html.DIV(Class="ui-progressbar-value ui-widget-header ui-corner-left",
                         style={'width': '0px'})
      self._div_shell <= self._bar

  def set_progress(self, percent):
      self._bar.style.width='%s%%' % percent
      if self._show_label:
         self._label.text='%s%%' % percent



# Module for profiling brython code.
#
# Written by Jonathan L. Verner
#
# Interface tries to follow the official python profile module by Sjoerd Mullender...
# which was hacked somewhat by: Guido van Rossum

"""
    Module for profiling Python code run in the browser via Brython.

    The module tries to provide interface which is as similar to the official python
    profile module. The notable difference is that it does not allow user-defined
    timers and does not do any callibration. Methods which in the standard module
    save the data to a file save a JSON-serialized version of the data to the browser's
    local storage instead.

    Basic usage:

        from profile import Profile

        p = Profile()
        p.enable()
        do_something()
        do_something_else()
        p.create_stats()

    Which will print out something like:

            1 run in 0.249 seconds

        Ordered by: standard name (averaged over 1 run)

        ncalls  tottime  percall  cumtime  var percall  module.function:lineno
         101/1    0.023    0.000    1.012        0.010               .fact:180

    where each line corresponds to a function and the different columns correspond  to

        ncalls      is the total number number of times the function was called
                    (if the function was called non-recursively, the second number
                    behind the backslash indicates how many calls were top-level calls
                    in the recursion)

        tottime     is the total time (in seconds) spent in the function not including subcalls

        percall     is the average time spent in the function per call, not including subcalls

        cumtime     is the total time spent in function including subcalls

        var percall is the average time spent in function per one non-recursive call

        standard name is the name of the function in the form module.function_name:line_number

    Optionally one can also use the following form, taking advantage of running the code
    several times and averaging it out:

        from profile import Profile

        p = Profile()
        p.call(function_to_profile,200,arg1,arg2,kwarg1=v1)

    which will print out something like:

            200 runs in 0.249 seconds

        Ordered by: standard name (averaged over 1 run)

        ncalls  tottime  percall  cumtime  var percall  module.function:lineno
         101/1    0.023    0.000    1.012        0.010  function_to_profile:16

    Collected profile data can be saved to local storage for later use:

        p.dump_stats('run1')

    Profile data can also be read back:

        data = Stats('run1')

    And aggregated together

        data.add(p.dump_stats())
        print(data)
"""

import _profile
import json
from browser.local_storage import storage

__all__ = ["run","runctx","status","Stats","Profile"]

class Stats:
    @classmethod
    def current_stats(cls):
        return Stats(_profile.data,1)

    def __init__(self,data,nruns=None):
        """
           If data is a string, it should be a JSON-serialized version of Stats.

           Otherwise it should be a javascript object containing the collected profile data.
           This other form is not meant for public use.
        """
        self._ordering = "standard name"
        self._reverse_order = False
        self._sk = None

        if isinstance(data,str):
            self.from_json(data)
        else:
            self._data = data
            self.nruns = nruns
            self.line_counts = {k:data.line_counts[k] for k in dir(data.line_counts)}
            self.duration = self._data.profile_duration
            self.function_totals = {k:data.call_times_proper[k] for k in dir(data.call_times_proper)}
            self.function_cumulated_totals = {k:data.call_times[k] for k in dir(data.call_times)}
            self.function_counts = {k:data.call_counts[k] for k in dir(data.call_counts)}
            self.function_counts_nonrec = {k:data.call_counts_norec[k] for k in dir(data.call_counts)}
            self.callers = {}
            for func in dir(data.callers):
                func_data = data.callers[func]
                fd = {}
                for caller in dir(func_data):
                    cumulated,total,count,count_norec = func_data[caller]
                    fd[caller] = {'cumulated':cumulated,'total':total,'count':count,'count_norec':count_norec}
                self.callers[func] = fd

    def __json__(self):
        return json.dumps({
            'nruns':self.nruns,
            'line_counts':self.line_counts,
            'duration':self.duration,
            'function_counts':self.function_counts,
            'function_counts_nonrec':self.function_counts_nonrec,
            'function_cumulated_totals':self.function_cumulated_totals,
            'function_totals':self.function_totals,
            'callers':self.callers
        })

    def from_json(self,json_data):
        self._data=None
        data = json.loads(json_data)
        self.nruns = data['nruns']
        self.line_counts = data['line_counts']
        self.duration = data['duration']
        self.function_counts = data['function_counts']
        self.function_counts_nonrec = data['function_counts_nonrec']
        self.function_cumulated_totals = data['function_cumulated_totals']
        self.function_totals = data['function_totals']
        self.callers = data['callers']


    def add(self,*stats):
        """
            Add profile data from other Stats objects or from
            Stats objects serialized into JSON and saved under
            in browser's local storage under the given keys.

            @stats should be a sequence of items where each item is
            either (a) Stats instance (b) a key under which a Stats
            instance is saved in JSON-serialized form in local storage
            (c) a JSON-serialized instance of a Stats object.
        """
        for s in stats:
            if isinstance(s,Stats):
                self._add_stat(s)
            else:
                if s in storage:
                    st = Stats(storage[s])
                else:
                    st = Stats(s)
                self._add_stat(st)

    def _add_stat(self,stat):
        for f in stat.function_totals.keys():
            if f not in self.function_counts.keys():
                self.function_totals[f] = 0
                self.function_cumulated_totals[f]=0
                self.function_counts[f] = 0
                self.function_counts_nonrec[f] = 0
            self.function_totals[f] += stat.function_totals[f]
            self.function_cumulated_totals[f] += stat.function_cumulated_totals[f]
            self.function_counts[f] += stat.function_counts[f]
            self.function_counts_nonrec[f] += stat.function_counts_nonrec[f]
        for (f,callers) in stat.callers.items():
            if f not in self.callers:
                sc = self.callers[f] = {}
            for (caller,values) in callers.items():
                if caller not in sc:
                    sc[caller] = {'cumulated':0,'total':0,'count':0,'count_norec':0}
                sc[caller] = { k:sc[caller][k]+values[k] for k in ['cumulated','total','count','count_norec'] }
        if self._sk is not None:
            self.sort()


    def sort(self,key='standard name'):
        """
            Sort the stats according to key. Key can be any of the columns
            in the table resulting from converting Stats to a string, i.e.:

             - standard name (module.function:line_number)
             - ncalls        (number of calls the function is called)
             - tottime       (total time spent in function not including subcalls)
             - percall       (average time spent per call, not including subcalls)
             - cumtime       (total time spent in function including subcalls)
             - var percall   (average time spent in function per one non-recursive call)

        """
        self._ordering = key
        self._sk = self._sorted_keys()

    def _sorted_keys(self):
        if self._ordering == 'standard name':
            return sorted(self.function_totals.keys())
        if self._ordering == 'ncalls':
            keys = self.function_counts.items()
        elif self._ordering == 'tottime':
            keys = self.function_totals.items()
        elif self._ordering == 'percall':
            keys = [ (k,self.function_totals[k]/self.function_counts[k]) for k in self.function_counts.keys() ]
        elif self._ordering == 'cumtime':
            keys = self.function_cumulated_totals.items()
        elif self._ordering == 'var percall':
            keys = [ (k,self.function_cumulated_totals[k]/self.function_counts_nonrec[k]) for k in self.function_counts_nonrec.keys() ]
        return [k for k,v in sorted(keys,key=lambda x:x[1], reverse=not self._reverse_order)]

    def __str__(self):
        if self.nruns == 1:
            nruns = " 1 run"
        else:
            nruns = " "+str(self.nruns)+" runs"

        summary = " "*8+nruns+" in " + str(self.duration/1000) + " seconds"
        headers = [" ncalls"," tottime"," percall"," cumtime", " var percall", " module.function:lineno"]
        cols = []
        if self._sk is None:
            self.sort()
        for func in self._sk:
            ncalls = "{:.0f}".format(self.function_counts[func]/self.nruns)
            if self.function_counts[func] > self.function_counts_nonrec[func]:
                ncalls +="/{:.0f}".format(self.function_counts_nonrec[func]/self.nruns)
            tottime = self.function_totals[func]/self.nruns
            percall = (tottime/self.function_counts[func])/self.nruns
            cumtime = self.function_cumulated_totals[func]/self.nruns
            varpercall = (cumtime/self.function_counts_nonrec[func])/self.nruns
            cols.append((ncalls,"{:.3f}".format(tottime/1000),"{:.3f}".format(percall/1000),"{:.3f}".format(cumtime/1000),"{:.3f}".format(varpercall/1000),func))
        header = ' '.join(headers)
        table = [summary,"","Ordered by: "+self._ordering+" (averaged over "+nruns+")","",header]
        for col in cols:
            ln = []
            for pos in range(len(headers)):
                ln.append(col[pos].rjust(len(headers[pos])))
            table.append(' '.join(ln))
        return "\n".join(table)

    def __repr__(self):
        return self.__str__()

class Profile:
    def __init__(self):
        self._profile = None

    def enable(self):
        """
            Start collecting profile data.
        """
        _profile.start()

    def disable(self):
        """
            Stop collecting profile data.
        """
        _profile.pause()

    def create_stats(self):
        """
            Stop collecting profile data and load them into
            a Stats object (which is returned).
        """
        _profile.stop()
        self._profile = Stats(_profile.data,1)
        return self._profile

    def status(self):
        return _profile.status

    def print_stats(self,sort='standard name'):
        """
            Print stats ordered according to the sort key @sort:
            The key can be any of the columns in the table
            which is printed out, i.e.:

             - standard name (module.function:line_number)
             - ncalls        (number of calls the function is called)
             - tottime       (total time spent in function not including subcalls)
             - percall       (average time spent per call, not including subcalls)
             - cumtime       (total time spent in function including subcalls)
             - var percall   (average time spent in function per one non-recursive call)

        """
        if self._profile is None:
            print("No profile available. Need to collect profile data first using the enable() and disable() methods.")
        else:
            self._profile.sort(sort)
            print(str(self._profile))

    def dump_stats(self,storage_key=None):
        """
            Saves a JSON-serialized version of the Stats object
            to local storage under the @storage_key key.
        """
        if storage_key is not None:
            storage[storage_key]=self._profile.__json__()
        else:
            return self._profile

    def run(self,cmd,nruns=200):
        """
            Runs the command @cmd and collects profile data.
            The command is run @ntimes (200 by default) and
            the data is averaged over these runs.
        """
        self.runctx(cmd,globals(),locals(),ntimes=nruns)

    def runctx(self,cmd,globals,locals,nruns=200):
        """
            Runs the command @cmd with the given @globlas and @locals
            and collects profile data. The command is run @ntimes (200 by default) and
            the data is averaged over these runs.
        """
        _profile.stop()
        _profile.clear()
        _profile.start()
        for i in range(nruns):
            _profile.run(cmd,globals=globals,locals=locals)
        _profile.stop()
        self._profile = Stats(_profile.data,nruns)

    def runcall(self,func,ntimes,*args,**kwargs):
        """
            Calls the function @func with arguments @args, @kwargs
            and collects profile data. The function is called @ntimes
            and the data is averaged over these runs.
        """
        _profile.stop()
        _profile.clear()
        _profile.start()
        for i in range(ntimes):
            ret=func(*args,**kwargs)
        _profile.stop()
        self._profile = Stats(_profile.data,ntimes)
        return self._profile

def run(cmd,nruns=200):
    """
        Runs the command @cmd and collects profile data.
        The command is run @ntimes (200 by default) and
        the data is averaged over these runs. Returns
        a Stats object with the collected data.
    """
    return runctx(cmd,globals(),locals(),nruns=nruns)

def runctx(cmd,globals=None,locals=None,nruns=200):
    """
        Runs the command @cmd with with the given @globlas and @locals
        and collects profile data. The command is run @nruns (200 by default) and
        the data is averaged over these runs. Returns
        a Stats object with the collected data.
    """
    _locals = locals or {}
    _globals = globals or {}
    _profile.start()
    _profile.run(cmd,_globals,_locals,nruns)
    _profile.stop()
    return Stats(_profile.data,nruns)

status = _profile.status


from . import widget
from browser import html, document

class Menu(html.UL, widget.Widget):
    def __init__(self, id=None, style={}):
        default_style = dict(position= 'relative', height='auto',
            width='auto')
        default_style.update(style)
        html.UL.__init__(self, Class="ui-widget ui-menu", style=default_style)
    
        widget.Widget.__init__(self, self, 'menu', id)
        document.bind('click', self.leave)
        self.active = False

    def add(self, title):
        item = MenuItem(title, Class="ui-widget ui-menu-item")
        item.bind('click', item.activate)
        item.bind('mouseenter', item.enter)
        self <= item
        return item

    def leave(self, ev):
        for child in self.children:
            if child.state == 'show':
                document.remove(child.div)
                child.state = 'hide'
        self.active = False

class MenuItem(html.LI):

    def __init__(self, *args, **kw):
        html.LI.__init__(self, *args, **kw)
        self.items = []
        self.state = "hide"

    def activate(self, ev):
        self.parent.active = True
        self.show(ev)
        ev.stopPropagation()

    def enter(self, ev):
        if self.parent.active:
            self.show(ev)

    def show(self, ev):
        for item in self.parent.children:
            if item.state == "show":
                if item == self:
                    return
                document.remove(item.div)
                item.state = "hide"
        if self.state == "hide":
            left = ev.target.left
            top = ev.target.top+ev.target.height

            self.div = html.DIV(Class="ui-widget ui-menu-sublist",
                style=dict(position='absolute', left=left, top=top, zIndex=99))
            for item in self.items:
                line = html.DIV(item[0], Class="ui-menu-subitem")
                if item[1] is not None:
                    line.bind('click', item[1])
                self.div <= line
            self.state = "show"
            self.div.style.borderWidth = "1px"
            document <= self.div
        else:
            document.remove(self.div)
            self.state = "hide"

    def add(self, label, callback = None):
        self.items.append((label, callback))



"""The machinery of importlib: finders, loaders, hooks, etc."""

import _imp

from ._bootstrap import (SOURCE_SUFFIXES, DEBUG_BYTECODE_SUFFIXES,
                         OPTIMIZED_BYTECODE_SUFFIXES, #BYTECODE_SUFFIXES,
                         EXTENSION_SUFFIXES)
from ._bootstrap import BuiltinImporter
from ._bootstrap import FrozenImporter
from ._bootstrap import WindowsRegistryFinder
from ._bootstrap import PathFinder
from ._bootstrap import FileFinder
from ._bootstrap import SourceFileLoader
from ._bootstrap import SourcelessFileLoader
from ._bootstrap import ExtensionFileLoader


#def all_suffixes():
#    """Returns a list of all recognized module suffixes for this process"""
#    return SOURCE_SUFFIXES + BYTECODE_SUFFIXES + EXTENSION_SUFFIXES



#!/usr/bin/env python3
"""
Operator Interface

This module exports a set of functions corresponding to the intrinsic 
operators of Python.  For example, operator.add(x, y) is equivalent 
to the expression x+y.  The function names are those used for special 
methods; variants without leading and trailing '__' are also provided 
for convenience.

This is the pure Python implementation of the module.
"""

# downloaded from http://bugs.python.org/file28327/operator.py

#import builtins as _bi  #there is no builtins module

def lt(a, b):
    "Same as a < b."
    return a < b
__lt__ = lt

def le(a, b):
    "Same as a <= b."
    return a <= b
__le__ = le

def eq(a, b):
    "Same as a == b."
    return a == b
__eq__ = eq

def ne(a, b):
    "Same as a != b."
    return a != b
__ne__ = ne

def ge(a, b):
    "Same as a >= b."
    return a >= b
__ge__ = ge

def gt(a, b):
    "Same as a > b."
    return a > b
__gt__ = gt

def not_(a):
    "Same as not a."
    return not a
__not__ = not_

def truth(a):
    "Return True if a is true, False otherwise."
    #return _bi.bool(a)
    return bool(a)

def is_(a, b):
    "Same as a is b."
    return a is b

# brython does not like  (causes syntax error)
#def is_not(a, b):
#    "Same as a is not b."
#    return a is not b

#recursion error or just comment out and add code below function
#def abs(a):
#    "Same as abs(a)."
#    #return _bi.abs(a)
#    return abs(a)
__abs__ = abs
abs=abs


def add(a, b):
    "Same as a + b."
    return a + b
__add__ = add

def and_(a, b):
    "Same as a & b."
    return a & b
__and__ = and_

def floordiv(a, b):
    "Same as a // b."
    return a // b
__floordiv__ = floordiv

def index(a):
    "Same as a.__index__()."
    return a.__index__()
__index__ = index

def inv(a):
    "Same as ~a."
    return ~a    #brython does not like
    #return a^(2**31)
invert = __inv__ = __invert__ = inv

def lshift(a, b):
    "Same as a << b."
    return a << b
__lshift__ = lshift

def mod(a, b):
    "Same as a % b."
    return a % b
__mod__ = mod

def mul(a, b):
    "Same as a * b."
    return a * b
__mul__ = mul

def neg(a):
    "Same as -a."
    return -a
__neg__ = neg

def or_(a, b):
    "Same as a | b."
    return a | b
__or__ = or_

def pos(a):
    "Same as +a."
    return +a    #brython does not like

__pos__ = pos

def pow(a, b):
    "Same as a ** b."
    return a ** b
__pow__ = pow

def rshift(a, b):
    "Same as a >> b."
    return a >> b
__rshift__ = rshift

def sub(a, b):
    "Same as a - b."
    return a - b
__sub__ = sub

def truediv(a, b):
    "Same as a / b."
    return a / b
__truediv__ = truediv

def xor(a, b):
    "Same as a ^ b."
    return a ^ b
__xor__ = xor

def concat(a, b):
    "Same as a + b, for a and b sequences."
    if not (hasattr(a, '__getitem__') and hasattr(b, '__getitem__')):
        raise TypeError('a and b must be sequences')
    return a + b
__concat__ = concat

def contains(a, b):
    "Same as b in a (note reversed operands)."
    return b in a
__contains__ = contains

def countOf(a, b):
    "Return the number of times b occurs in a."
    count = 0
    for i in a:
        if i == b:
            count += 1
    return count

def delitem(a, b):
    "Same as del a[b]."
    del a[b]
__delitem__ = delitem

def getitem(a, b):
    "Same as a[b]."
    return a[b]
__getitem__ = getitem

#fixme  brython doesn't like this function
def indexOf(a, b):
    "Return the first index of b in a."
    #for i, j in _bi.enumerate(a):
    for i, j in enumerate(a):
        if j == b:
            return i
    else:
        raise ValueError('b not found in a')

def setitem(a, b, c):
    "Same as a[b] = c."
    a[b] = c
__setitem__ = setitem



class attrgetter:
    """
    Return a callable object that fetches the given attribute(s) from its operand.
    After f=attrgetter('name'), the call f(r) returns r.name.
    After g=attrgetter('name', 'date'), the call g(r) returns (r.name, r.date).
    After h=attrgetter('name.first', 'name.last'), the call h(r) returns
    (r.name.first, r.name.last).
    """
    def __init__(self, attr, *attrs):
        self._attrs = (attr,)
        self._attrs += attrs
        if any(not isinstance(attr, str) for attr in self._attrs):
            raise TypeError('attribute name must be a string')

    @staticmethod
    def _resolve_attr(obj, attr):
        for name in attr.split('.'):
            #obj = _bi.getattr(obj, name)
            obj = getattr(obj, name)
        return obj

    def __call__(self, obj):
        if len(self._attrs) == 1:
            return self._resolve_attr(obj, self._attrs[0])
        return tuple(self._resolve_attr(obj, attr) for attr in self._attrs)

class itemgetter:
    """
    Return a callable object that fetches the given item(s) from its operand.
    After f=itemgetter(2), the call f(r) returns r[2].
    After g=itemgetter(2,5,3), the call g(r) returns (r[2], r[5], r[3])
    """
    def __init__(self, item, *items):
        self._items = (item,)
        self._items += items

    def __call__(self, obj):
        if len(self._items) == 1:
            return obj[self._items[0]]
        return tuple(obj[item] for item in self._items)

class methodcaller:
    """
    Return a callable object that calls the given method on its operand.
    After f = methodcaller('name'), the call f(r) returns r.name().
    After g = methodcaller('name', 'date', foo=1), the call g(r) returns
    r.name('date', foo=1).
    """

    def __init__(self, name, *args, **kwargs):
        self._name = name
        self._args = args
        self._kwargs = kwargs

    def __call__(self, obj):
        return getattr(obj, self._name)(*self._args, **self._kwargs)


def iadd(a, b):
    "Same as a += b."
    a += b
    return a
__iadd__ = iadd

def iand(a, b):
    "Same as a &= b."
    a &= b
    return a
__iand__ = iand

def iconcat(a, b):
    "Same as a += b, for a and b sequences."
    if not (hasattr(a, '__getitem__') and hasattr(b, '__getitem__')):
        raise TypeError('a and b must be sequences')
    a += b
    return a
__iconcat__ = iconcat

def ifloordiv(a, b):
    "Same as a //= b."
    a //= b
    return a
__ifloordiv__ = ifloordiv

def ilshift(a, b):
    "Same as a <<= b."
    a <<= b
    return a
__ilshift__ = ilshift

def imod(a, b):
    "Same as a %= b."
    a %= b
    return a
__imod__ = imod

def imul(a, b):
    "Same as a *= b."
    a *= b
    return a
__imul__ = imul

def ior(a, b):
    "Same as a |= b."
    a |= b
    return a
__ior__ = ior

def ipow(a, b):
    "Same as a **= b."
    a **=b
    return a
__ipow__ = ipow

def irshift(a, b):
    "Same as a >>= b."
    a >>= b
    return a
__irshift__ = irshift

def isub(a, b):
    "Same as a -= b."
    a -= b
    return a
__isub__ = isub

def itruediv(a, b):
    "Same as a /= b."
    a /= b
    return a
__itruediv__ = itruediv

def ixor(a, b):
    "Same as a ^= b."
    a ^= b
    return a
__ixor__ = ixor

def length_hint(obj, default=0):
    """
    Return an estimate of the number of items in obj.
    This is useful for presizing containers when building from an iterable.

    If the object supports len(), the result will be exact. Otherwise, it may
    over- or under-estimate by an arbitrary amount. The result will be an
    integer >= 0.
    """
    try:
        return len(obj)
    except TypeError:
        try:
            val = obj.__length_hint__()
            if val is NotImplemented:
                raise TypeError
        except (AttributeError, TypeError):
            return default
        else:
            if not val > 0:
                raise ValueError('default must be > 0')
            return val

#try:
#    from _operator import *
#    from _operator import __doc__
#except ImportError:
#   pass



import operator

class accumulate:
    def __init__(self, iterable, func = operator.add):
        self.it = iter(iterable)
        self._total = None
        self.func = func
        
    def __iter__(self):
        return self

    def __next__(self):
        if not self._total:
            self._total = next(self.it)
            return self._total
        else:
            element = next(self.it)
            try:
                self._total = self.func(self._total, element)
            except:
                raise TypeError("unsupported operand type")
            return self._total
                
## Adapted from:
## https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-34
class chain:
    def __init__(self, *iterables):
        self._iterables_iter = iter(map(iter, iterables))
        # little trick for the first chain.__next__() call
        self._cur_iterable_iter = iter([])

    def __iter__(self):
        return self
    
    def __next__(self):
        while True:
            try:
                return next(self._cur_iterable_iter)
            except StopIteration:
                self._cur_iterable_iter = next(self._iterables_iter)
    
    @classmethod
    def from_iterable(cls, iterable):
        for it in iterable:
            for element in it:
                yield element
                
class combinations:
    def __init__(self, iterable, r):
        self.pool = tuple(iterable)
        self.n = len(self.pool)
        self.r = r
        self.indices = list(range(self.r))
        self.zero = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.r > self.n:
            raise StopIteration
        if not self.zero:
            self.zero = True
            return tuple(self.pool[i] for i in self.indices)
        else:
            try:
                for i in reversed(range(self.r)):
                    if self.indices[i] != i + self.n - self.r:
                        break
                self.indices[i] += 1
                for j in range(i+1, self.r):
                    self.indices[j] = self.indices[j-1] + 1
                return tuple(self.pool[i] for i in self.indices)
            except:
                raise StopIteration
                
class combinations_with_replacement:
    def __init__(self, iterable, r):
        self.pool = tuple(iterable)
        self.n = len(self.pool)
        self.r = r
        self.indices = [0] * self.r
        self.zero = False
        
    def __iter__(self):
        return self
    
    def __next__(self):
        if not self.n and self.r:
            raise StopIteration
        if not self.zero:
            self.zero = True
            return tuple(self.pool[i] for i in self.indices)
        else:
            try:
                for i in reversed(range(self.r)):
                    if self.indices[i] != self.n - 1:
                        break
                self.indices[i:] = [self.indices[i] + 1] * (self.r - i)
                return tuple(self.pool[i] for i in self.indices)
            except:
                raise StopIteration
                
## Literally copied from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-63 
class compress:
    def __init__(self, data, selectors):
        self.data = iter(data)
        self.selectors = iter(selectors)

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            next_item = next(self.data)
            next_selector = next(self.selectors)
            if bool(next_selector):
                return next_item
                
## Adapted from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-79
## I mimicked the > python3.1 behavior
class count:
    """
    Input is an int or a float. The original Python 3 implementation
    includes also complex numbers... but it still is not implemented 
    in Brython as complex type is NotImplemented
    """
    def __init__(self, start = 0, step = 1):
        if not isinstance(start, (int, float)):
            raise TypeError('a number is required')
        self.times = start - step
        self.step = step

    def __iter__(self):
        return self

    def __next__(self):
        self.times += self.step
        return self.times

    def __repr__(self):
        return 'count(%d)' % (self.times + self.step)

## Literally copied from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-112        
class cycle:
    def __init__(self, iterable):
        self._cur_iter = iter(iterable)
        self._saved = []
        self._must_save = True
        
    def __iter__(self):
        return self

    def __next__(self):
        try:
            next_elt = next(self._cur_iter)
            if self._must_save:
                self._saved.append(next_elt)
        except StopIteration:
            self._cur_iter = iter(self._saved)
            next_elt = next(self._cur_iter)
            self._must_save = False
        return next_elt
        
## Literally copied from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-149
class dropwhile:
    def __init__(self, predicate, iterable):
        self._predicate = predicate
        self._iter = iter(iterable)
        self._dropped = False

    def __iter__(self):
        return self

    def __next__(self):
        value = next(self._iter)
        if self._dropped:
            return value
        while self._predicate(value):
            value = next(self._iter)
        self._dropped = True
        return value
        
## Adapted from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-261
class filterfalse:
    def __init__(self, predicate, iterable):
        # Make sure iterable *IS* iterable
        self._iter = iter(iterable)
        if predicate is None:
            self._predicate = bool
        else:
            self._predicate = predicate

    def __iter__(self):
        return self
    def __next__(self):
        next_elt = next(self._iter)
        while True:
            if not self._predicate(next_elt):
                return next_elt
            next_elt = next(self._iter)

class groupby:
    # [k for k, g in groupby('AAAABBBCCDAABBB')] --> A B C D A B
    # [list(g) for k, g in groupby('AAAABBBCCD')] --> AAAA BBB CC D
    def __init__(self, iterable, key=None):
        if key is None:
            key = lambda x: x
        self.keyfunc = key
        self.it = iter(iterable)
        self.tgtkey = self.currkey = self.currvalue = object()
    def __iter__(self):
        return self
    def __next__(self):
        while self.currkey == self.tgtkey:
            self.currvalue = next(self.it)    # Exit on StopIteration
            self.currkey = self.keyfunc(self.currvalue)
        self.tgtkey = self.currkey
        return (self.currkey, self._grouper(self.tgtkey))
    def _grouper(self, tgtkey):
        while self.currkey == tgtkey:
            yield self.currvalue
            self.currvalue = next(self.it)    # Exit on StopIteration
            self.currkey = self.keyfunc(self.currvalue)
          
## adapted from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-323
class islice:
    def __init__(self, iterable, *args):
        s = slice(*args)
        self.start, self.stop, self.step = s.start or 0, s.stop, s.step
        if not isinstance(self.start, int):
           raise ValueError("Start argument must be an integer")
        if self.stop != None and not isinstance(self.stop, int):
            raise ValueError("Stop argument must be an integer or None")
        if self.step is None:
            self.step = 1
        if self.start<0 or (self.stop != None and self.stop<0
           ) or self.step<=0:
            raise ValueError("indices for islice() must be positive")
        self.it = iter(iterable)
        self.donext = None
        self.cnt = 0

    def __iter__(self):
        return self

    def __next__(self):
        nextindex = self.start
        if self.stop != None and nextindex >= self.stop:
            raise StopIteration
        while self.cnt <= nextindex:
            nextitem = next(self.it)
            self.cnt += 1
        self.start += self.step 
        return nextitem
        
class permutations:
    def __init__(self, iterable, r = None):
        self.pool = tuple(iterable)
        self.n = len(self.pool)
        self.r = self.n if r is None else r
        self.indices = list(range(self.n))
        self.cycles = list(range(self.n, self.n - self.r, -1))
        self.zero = False
        self.stop = False

    def __iter__(self):
        return self

    def __next__(self):
        indices = self.indices
        if self.r > self.n:
            raise StopIteration
        if not self.zero:
            self.zero = True
            return tuple(self.pool[i] for i in indices[:self.r])
        
        i = self.r - 1
        while i >= 0:
            j = self.cycles[i] - 1
            if j > 0:
                self.cycles[i] = j
                indices[i], indices[-j] = indices[-j], indices[i]
                return tuple(self.pool[i] for i in indices[:self.r])
            self.cycles[i] = len(indices) - i
            n1 = len(indices) - 1
            assert n1 >= 0
            num = indices[i]
            for k in range(i, n1):
                indices[k] = indices[k+1]
            indices[n1] = num
            i -= 1
        raise StopIteration

# copied from Python documentation on itertools.product
def product(*args, repeat=1):
    # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
    # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
    pools = [tuple(pool) for pool in args] * repeat
    result = [[]]
    for pool in pools:
        result = [x+[y] for x in result for y in pool]
    for prod in result:
        yield tuple(prod)


## adapted from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-392        

## (Brython) 
## renamed to _product : the implementation fails for product('abc', [])
## with CPython 3.x
class _product:
    def __init__(self, *args, **kw):
        if len(kw) > 1:
            raise TypeError("product() takes at most 1 argument (%d given)" %
                             len(kw))
        self.repeat = kw.get('repeat', 1)
        if not isinstance(self.repeat, int):
            raise TypeError("integer argument expected, got %s" %
                             type(self.repeat))
        self.gears = [x for x in args] * self.repeat
        self.num_gears = len(self.gears)
        # initialization of indicies to loop over
        self.indicies = [(0, len(self.gears[x]))
                         for x in range(0, self.num_gears)]
        self.cont = True
        self.zero = False

    def roll_gears(self):
        # Starting from the end of the gear indicies work to the front
        # incrementing the gear until the limit is reached. When the limit
        # is reached carry operation to the next gear
        should_carry = True
        for n in range(0, self.num_gears):
            nth_gear = self.num_gears - n - 1
            if should_carry:
                count, lim = self.indicies[nth_gear]
                count += 1
                if count == lim and nth_gear == 0:
                    self.cont = False
                if count == lim:
                    should_carry = True
                    count = 0
                else:
                    should_carry = False
                self.indicies[nth_gear] = (count, lim)  
            else:
                break

    def __iter__(self):
        return self

    def __next__(self):
        if self.zero:
            raise StopIteration
        if self.repeat > 0:
            if not self.cont:
                raise StopIteration
            l = []
            for x in range(0, self.num_gears):
                index, limit = self.indicies[x]
                print('itertools 353',self.gears,x,index)
                l.append(self.gears[x][index])
            self.roll_gears()
            return tuple(l)
        elif self.repeat == 0:
            self.zero = True
            return ()
        else:
            raise ValueError("repeat argument cannot be negative")
            
## Literally copied from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-441
class repeat:
    def __init__(self, obj, times=None):
        self._obj = obj
        if times is not None:
            range(times) # Raise a TypeError
            if times < 0:
                times = 0
        self._times = times
        
    def __iter__(self):
        return self

    def __next__(self):
        # __next__() *need* to decrement self._times when consumed
        if self._times is not None:
            if self._times <= 0: 
                raise StopIteration()
            self._times -= 1
        return self._obj

    def __repr__(self):
        if self._times is not None:
            return 'repeat(%r, %r)' % (self._obj, self._times)
        else:
            return 'repeat(%r)' % (self._obj,)

    def __len__(self):
        if self._times == -1 or self._times is None:
            raise TypeError("len() of uniszed object")
        return self._times

## Adapted from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-489
class starmap(object):
    def __init__(self, function, iterable):
        self._func = function
        self._iter = iter(iterable)

    def __iter__(self):
        return self

    def __next__(self):
        t = next(self._iter)
        return self._func(*t)
        
## Literally copied from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-520
class takewhile(object):
    def __init__(self, predicate, iterable):
        self._predicate = predicate
        self._iter = iter(iterable)

    def __iter__(self):
        return self

    def __next__(self):
        value = next(self._iter)
        if not self._predicate(value):
            raise StopIteration()
        return value

## Almost literal from
##https://bitbucket.org/pypy/pypy/src/c1aa74c06e86/lib_pypy/itertools.py#cl-547
class TeeData(object):
    def __init__(self, iterator):
        self.data = []
        self._iter = iterator

    def __getitem__(self, i):
        # iterates until 'i' if not done yet
        while i>= len(self.data):
            self.data.append(next(self._iter))
        return self.data[i]


class TeeObject(object):
    def __init__(self, iterable=None, tee_data=None):
        if tee_data:
            self.tee_data = tee_data
            self.pos = 0
        # <=> Copy constructor
        elif isinstance(iterable, TeeObject):
            self.tee_data = iterable.tee_data
            self.pos = iterable.pos
        else:
            self.tee_data = TeeData(iter(iterable))
            self.pos = 0
            
    def __next__(self):
        data = self.tee_data[self.pos]
        self.pos += 1
        return data
    
    def __iter__(self):
        return self


def tee(iterable, n=2):
    if isinstance(iterable, TeeObject):
        return tuple([iterable] +
        [TeeObject(tee_data=iterable.tee_data) for i in range(n - 1)])
    tee_data = TeeData(iter(iterable))
    return tuple([TeeObject(tee_data=tee_data) for i in range(n)])

class zip_longest:
    def __init__(self, *args, fillvalue = None):
        self.args = [iter(arg) for arg in args]
        self.fillvalue = fillvalue
        self.units = len(args)
    
    def __iter__(self):
        return self
    
    def __next__(self):
        temp = []
        nb = 0
        for i in range(self.units):
            try:
                temp.append(next(self.args[i]))
                nb += 1
            except StopIteration:
                temp.append(self.fillvalue)
        if nb==0:
            raise StopIteration
        return tuple(temp)



# Import modules
from collections import OrderedDict


MESSAGES = {
    'default': 'Islet Regulome intro page',
    'not_implemented': 'Not implemented yet'
}

SNPS = OrderedDict([
    ('diagram', 'T2D - DIAGRAM'),
    ('magic', 'Fasting glycemia – MAGIC')
])

BUILDS = OrderedDict([
    ('hg18', 'GRCh36/hg18'),
    ('hg19', 'GRCh37/hg19'),
    ('hg38', 'GRCh38/hg38')
])

REGIONS = OrderedDict([
    ('adult', 'Adult pancreatic islets - Open chromatin classes'),
    ('adultStretch', 'Adult pancreatic islets - Chromatin states'),
    ('progenitor', 'Pancreatic islet progenitors - Regulatory regions')
])

TFBS = OrderedDict([('adult', 'Adult pancreatic islets - TFs'),
                    ('progenitor', 'Pancreatic islet progenitors - TFs')
                    ])

CHROMATIN = OrderedDict([('alpha', 'FACS purified alpha cells – open chromatin profile'),
                         ('beta', 'FACS purified beta cells - open chromatin profile'),
                         ('none', 'None')
                         ])

RANGES = OrderedDict([(0, '0bp'),
                      (1000, '1Kb'),
                      (5000, '5Kb'),
                      (10000, '10Kb'),
                      (50000, '50Kb'),
                      (100000, '100Kb'),
                      (500000, '500Kb'),
                      (1000000, '1Mb')]
                     )
RANGES_R = {value: key for key, value in RANGES.items()}


CHROMOSOMES = [str(n) for n in range(1, 23)]
CHROMOSOMES.extend(["X", "Y"])
CHROMOSOMES_DICT = {'chromosome {}'.format(chromosome): 'chr{}'.format(chromosome)
                    for chromosome in CHROMOSOMES}
CHROMOSOMES_DICT_R = {value: key for key, value in CHROMOSOMES_DICT.items()}
CHROMOSOMES_DATASET = {'chromosome {}'.format(chromosome): '' for chromosome in CHROMOSOMES}
CHROMOSOMES_SORTED = ["chromosome {}".format(i) for i in CHROMOSOMES]
CHROMOSOME_LENGTHS = {
    "hg18": {
        'chromosome 1': 247249719,
        'chromosome 2': 242951149,
        'chromosome 3': 199501827,
        'chromosome 4': 191273063,
        'chromosome 5': 180857866,
        'chromosome 6': 170899992,
        'chromosome 7': 158821424,
        'chromosome 8': 146274826,
        'chromosome 9': 140273252,
        'chromosome 10': 135374737,
        'chromosome 11': 134452384,
        'chromosome 12': 132349534,
        'chromosome 13': 114142980,
        'chromosome 14': 106368585,
        'chromosome 15': 100338915,
        'chromosome 16': 88827254,
        'chromosome 17': 78774742,
        'chromosome 18': 76117153,
        'chromosome 19': 63811651,
        'chromosome 20': 62435964,
        'chromosome 21': 46944323,
        'chromosome 22': 49691432,
        'chromosome X': 154913754,
        'chromosome Y': 57772954,
        'chromosome M': 16571
    },
    "hg19": {
        'chromosome 1': 249250621,
        'chromosome 2': 243199373,
        'chromosome 3': 198022430,
        'chromosome 4': 191154276,
        'chromosome 5': 180915260,
        'chromosome 6': 171115067,
        'chromosome 7': 159138663,
        'chromosome 8': 146364022,
        'chromosome 9': 141213431,
        'chromosome 10': 135534747,
        'chromosome 11': 135006516,
        'chromosome 12': 133851895,
        'chromosome 13': 115169878,
        'chromosome 14': 107349540,
        'chromosome 15': 102531392,
        'chromosome 16': 90354753,
        'chromosome 17': 81195210,
        'chromosome 18': 78077248,
        'chromosome 19': 59128983,
        'chromosome 20': 63025520,
        'chromosome 21': 48129895,
        'chromosome 22': 51304566,
        'chromosome X': 155270560,
        'chromosome Y': 59373566,
        'chromosome M': 16571
    },
    "hg38": {
        'chromosome 1': 248956422,
        'chromosome 2': 242193529,
        'chromosome 3': 198295559,
        'chromosome 4': 190214555,
        'chromosome 5': 181538259,
        'chromosome 6': 170805979,
        'chromosome 7': 159345973,
        'chromosome 8': 145138636,
        'chromosome 9': 138394717,
        'chromosome 10': 133797422,
        'chromosome 11': 135086622,
        'chromosome 12': 133275309,
        'chromosome 13': 114364328,
        'chromosome 14': 107043718,
        'chromosome 15': 101991189,
        'chromosome 16': 90338345,
        'chromosome 17': 83257441,
        'chromosome 18': 80373285,
        'chromosome 19': 58617616,
        'chromosome 20': 64444167,
        'chromosome 21': 46709983,
        'chromosome 22': 50818468,
        'chromosome X': 156040895,
        'chromosome Y': 57227415,
        'chromosome M': 16569
    }
}

COLORS = {
    "adult": {
        "Promoter": (0, 0, 0),
        "Inactive enhancer": (171, 217, 233),
        "Active enhancer": (215, 25, 28),
        "C3 cluster": (152, 17, 53),
        "CTCF": (44, 123, 182),
        "Others": (253, 174, 97),
        "GENE_SPECIFIC": (100, 100, 100),
        "GENE_UNSPECIFIC": (200, 200, 200),
        "TFBS": (44, 65, 182)
    },
    "adultStretch": {
        "Insulator": (0, 0, 0),
        "Weak enhancer": (171, 217, 233),
        "Active promoter": (215, 25, 28),
        "Weak promoter": (44, 123, 182),
        "Strong enhancer": (253, 174, 97),
        "Repressed": (77, 175, 74),
        "Transcribed": (152, 78, 163),
        "Posied promoter": (166, 86, 40),
        "Stretch Enhancer": (152, 17, 53),
        "GENE_SPECIFIC": (100, 100, 100),
        "GENE_UNSPECIFIC": (200, 200, 200),
        "TFBS": (44, 65, 182)
    },
    "progenitor": {
        "Active enhancer": (0, 0, 0),
    }
}


# Table of colors
# name	        hex	    rgb	        adultIslet	        adultIsletStretch	progenitor
# ------------------------------------------------------------------------------------------
# black	        #000000	0,0,0	    Promoter	        Insulator	        Active enhancer
# light blue	#ABD9E9	171,217,233	Inactive enhancer	Weak enhancer	    -
# red	        #D7191C	215,25,28	Active enhancer	    Active promoter	    -
# blue	        #2C7BB6	44,123,182	CTCF	            Weak promoter	    -
# beige	        #FDAE61	253,174,97	Others	            Strong enhancer	    -
# green	        #4DAF4A	77,175,74	-	                Repressed	        -
# purple	    #984EA3	152,78,163	-	                Transcribed	        -
# brown	        #A65628	166,86,40	-	                Posied promoter	    -
# ------------------------------------------------------------------------------------------


default_messages = {
    'build': 'hg19',
    'region': 'adult',
    'tfbs': 'adult',
    'chromatin': 'none',
    'gene': '',
    'chromosome': 'chromosome 13',
    'chromosome_number': '13',
    'start': '',
    'end': '',
    'range_region': '50000',
    'select_snps': 'diagram_magic',
    'upload_regions': 'None',
    'upload_snps': 'None',
    'shared_regions': 'None',
    'shared_snps': 'None',
    'action': ''
}





from browser import window

__all__ = ["Error", "open", "open_new", "open_new_tab"]

class Error(Exception):
    pass

_target = { 0: '', 1: '_blank', 2: '_new' }  # hack...


def open(url, new=0, autoraise=True):
    """ 
    new window or tab is not controllable
    on the client side. autoraise not available.
    """
    # javascript window.open doesn't work if you do not specify the protocol
    # A solution is the next hack:
    if '://' in url:
        if url[:6] == 'ftp://':
            print('entro')
        else:
            protocol = url.split('//:')[0]
            url = url.replace(protocol + '//:', '//')
    else:
        url = '//' + url
    print(url)
    if window.open(url, _target[new]):
        return True
    return False

def open_new(url):
    return open(url, 1)

def open_new_tab(url):
    return open(url, 2)





from browser import document, window, alert


panels = {
    'data_human_islet': (
        'data_pancreatic_islet_enhancers',
        'data_human_b_cell',
        'data_open_chromatin',
        'data_chromatin_state',
        'data_human_alpha_cell'
    ),
    'data_progenitor': (
        'data_invitro_progenitors',
    ),
    'data_gwas': (
        'data_diagram',
        'data_magic'
    )

}


def toggle_panel(hash):
    """Open and close the main panels"""
    for k, v in panels.items():
        if k == hash:
            document[k].class_name = 'panel-collapse collapse in'
            document[k + "_icon"].class_name = 'heading-basic-icon'
        else:
            document[k].class_name = 'panel-collapse collapse'
            document[k + "_icon"].class_name = 'heading-basic-icon collapsed'


def toggle_subpanel(hash):
    """Open and close the main panels"""
    for k, v in panels.items():
        for i in v:
            if i == hash:
                document[i].class_name = 'panel-collapse collapse in'
                document[i + "_icon"].class_name = 'heading-submit-options-icon'
                toggle_panel(k)
            else:
                document[i].class_name = 'panel-collapse collapse'
                document[i + "_icon"].class_name = 'heading-submit-options-icon collapsed'


def collapse_in():
    """Open a collapsed panel"""
    hash = window.location.hash

    if hash != '':
        hash = hash[1:]

        for key, values in panels.items():
            if hash == key:
                toggle_panel(hash)
            elif hash in values:
                toggle_subpanel(hash)


collapse_in()



from browser import document as doc
from browser import window, html, markdown

import highlight

import time

def run(ev):
    # run the code in the elt after the button
    ix = ev.target.parent.children.index(ev.target)
    elt = ev.target.parent.children[ix+1]
    exec(elt.text)
    elt.focus()

def load(url,target):
    # fake query string to bypass browser cache
    qs = '?foo=%s' %time.time()
    try:
        mk,scripts = markdown.mark(open(url+qs).read())
    except IOError:
        doc[target].html = "Page %s not found" %url
        return False
    doc[target].html = mk
    for script in scripts:
        exec(script)
    for elt in doc[target].get(selector='.exec'):
        # Python code executed when user clicks on a button
        elt.contentEditable = True
        src = elt.text.strip()
        h = highlight.highlight(src)
        h.className = "pycode"
        elt.clear()
        elt <= h
        elt.focus()
        btn = html.BUTTON('▶')
        btn.bind('click', run)
        elt.parent.insertBefore(btn, elt)
    for elt in doc[target].get(selector='.exec_on_load'):
        # Python code executed on page load
        src = elt.text.strip()
        h = highlight.highlight(src)
        h.className = "pycode"
        elt.clear()
        elt <= h
        exec(src)
    for elt in doc[target].get(selector='.python'):
        src = elt.text.strip()
        h = highlight.highlight(src)
        h.className = "pycode"
        elt.clear()
        elt <= h
    return False



from browser import document, window, alert


panels = (
    'info_create_plot',
    'info_plot_description',
    'info_moving_around',
    'info_upload_your_data',
    'info_retrieve_results'
)


def toggle_panel(hash):
    """Open and close the main panels"""
    for panel in panels:
        if panel == hash:
            document[panel].class_name = 'panel-collapse collapse in'
            document[panel + "_icon"].class_name = 'heading-advanced-icon'
        else:
            document[panel].class_name = 'panel-collapse collapse'
            document[panel + "_icon"].class_name = 'heading-advanced-icon collapsed'


def collapse_in():
    """Open a collapsed panel"""
    hash = window.location.hash

    if hash != '':
        hash = hash[1:]

        for panel in panels:
            if hash == panel:
                toggle_panel(hash)


collapse_in()



import pickle

class __UnProvided():
    pass


class ObjectStorage():

    def __init__(self, storage):
        self.storage = storage

    def __delitem__(self, key):
        del self.storage[pickle.dumps(key)]

    def __getitem__(self, key):
        return pickle.loads(self.storage[pickle.dumps(key)])

    def __setitem__(self, key, value):
        self.storage[pickle.dumps(key)] = pickle.dumps(value)

    def __contains__(self, key):
        return pickle.dumps(key) in self.storage

    def get(self, key, default=None):
        if pickle.dumps(key) in self.storage:
            return self.storage[pickle.dumps(key)]
        return default

    def pop(self, key, default=__UnProvided()):
        if type(default) is __UnProvided or pickle.dumps(key) in self.storage:
            return pickle.loads(self.storage.pop(pickle.dumps(key)))
        return default

    def __iter__(self):
        keys = self.keys()
        return keys.__iter__()

    def keys(self):
        return [pickle.loads(key) for key in self.storage.keys()]

    def values(self):
        return [pickle.loads(val) for val in self.storage.values()]

    def items(self):
        return list(zip(self.keys(), self.values()))

    def clear(self):
        self.storage.clear()

    def __len__(self):
        return len(self.storage)



from browser import ajax
from . import error

class FileIO:
  def __init__(self, data):
      self._data=data

  def read(self):
      return self._data

def urlopen(url, data=None, timeout=None):
    global result
    result=None

    def on_complete(req):
        global result
        if req.status == 200:
            result=req

    _ajax=ajax.ajax()
    _ajax.bind('complete', on_complete)
    if timeout is not None:
       _ajax.set_timeout(timeout)

    if data is None:
       _ajax.open('GET', url, False)
       _ajax.send()
    else:
       _ajax.open('POST', url, False)
       _ajax.send(data)

    if result is not None:
        if isinstance(result.text, str):
           return FileIO(result.text) #, url, {'status': result.status}
    
        return FileIO(result.text()) #, url, {'status': result.status}
    raise error.HTTPError('file not found')



def ascii_decode(*args,**kw):
    pass

def ascii_encode(*args,**kw):
    pass

def charbuffer_encode(*args,**kw):
    pass

def charmap_build(*args,**kw):
    pass

def charmap_decode(*args,**kw):
    pass

def charmap_encode(*args,**kw):
    pass

def decode(*args,**kw):
    """decode(obj, [encoding[,errors]]) -> object    
    Decodes obj using the codec registered for encoding. encoding defaults
    to the default encoding. errors may be given to set a different error
    handling scheme. Default is 'strict' meaning that encoding errors raise
    a ValueError. Other possible values are 'ignore' and 'replace'
    as well as any other name registered with codecs.register_error that is
    able to handle ValueErrors."""
    pass

def encode(*args,**kw):
    """encode(obj, [encoding[,errors]]) -> object    
    Encodes obj using the codec registered for encoding. encoding defaults
    to the default encoding. errors may be given to set a different error
    handling scheme. Default is 'strict' meaning that encoding errors raise
    a ValueError. Other possible values are 'ignore', 'replace' and
    'xmlcharrefreplace' as well as any other name registered with
    codecs.register_error that can handle ValueErrors."""
    obj = args[0]
    if len(args)==2:
        encoding = args[1]
    else:
        encoding = 'utf-8'
    if isinstance(obj, str):
        return obj.encode(encoding)

def escape_decode(*args,**kw):
    pass

def escape_encode(*args,**kw):
    pass

def latin_1_decode(*args,**kw):
    pass

def latin_1_encode(*args,**kw):
    pass

def lookup(encoding):
    """lookup(encoding) -> CodecInfo    
    Looks up a codec tuple in the Python codec registry and returns
    a CodecInfo object."""

    if encoding in ('utf-8', 'utf_8'):
       from javascript import console
       console.log('encoding', encoding)
       import encodings.utf_8
       return encodings.utf_8.getregentry()

    LookupError(encoding)

def lookup_error(*args,**kw):
    """lookup_error(errors) -> handler    
    Return the error handler for the specified error handling name
    or raise a LookupError, if no handler exists under this name."""
    pass

def mbcs_decode(*args,**kw):
    pass

def mbcs_encode(*args,**kw):
    pass

def raw_unicode_escape_decode(*args,**kw):
    pass

def raw_unicode_escape_encode(*args,**kw):
    pass

def readbuffer_encode(*args,**kw):
    pass

def register(*args,**kw):
    """register(search_function)    
    Register a codec search function. Search functions are expected to take
    one argument, the encoding name in all lower case letters, and return
    a tuple of functions (encoder, decoder, stream_reader, stream_writer)
    (or a CodecInfo object)."""
    pass

def register_error(*args,**kw):
    """register_error(errors, handler)    
    Register the specified error handler under the name
    errors. handler must be a callable object, that
    will be called with an exception instance containing
    information about the location of the encoding/decoding
    error and must return a (replacement, new position) tuple."""
    pass

def unicode_escape_decode(*args,**kw):
    pass

def unicode_escape_encode(*args,**kw):
    pass

def unicode_internal_decode(*args,**kw):
    pass

def unicode_internal_encode(*args,**kw):
    pass

def utf_16_be_decode(*args,**kw):
    pass

def utf_16_be_encode(*args,**kw):
    pass

def utf_16_decode(*args,**kw):
    pass

def utf_16_encode(*args,**kw):
    pass

def utf_16_ex_decode(*args,**kw):
    pass

def utf_16_le_decode(*args,**kw):
    pass

def utf_16_le_encode(*args,**kw):
    pass

def utf_32_be_decode(*args,**kw):
    pass

def utf_32_be_encode(*args,**kw):
    pass

def utf_32_decode(*args,**kw):
    pass

def utf_32_encode(*args,**kw):
    pass

def utf_32_ex_decode(*args,**kw):
    pass

def utf_32_le_decode(*args,**kw):
    pass

def utf_32_le_encode(*args,**kw):
    pass

def utf_7_decode(*args,**kw):
    pass

def utf_7_encode(*args,**kw):
    pass

def utf_8_decode(*args,**kw):
    pass

def utf_8_encode(*args,**kw):
    input=args[0]
    if len(args) == 2:
       errors = args[1]
    else:
       errors=kw.get('errors', 'strict')

    #todo need to deal with errors, but for now assume all is well.

    return (bytes([_f for _f in input], 'utf-8'), len(input))



def partial(func, *args, **keywords):
    def newfunc(*fargs, **fkeywords):
        newkeywords = keywords.copy()
        newkeywords.update(fkeywords)
        return func(*(args + fargs), **newkeywords)
    newfunc.func = func
    newfunc.args = args
    newfunc.keywords = keywords
    return newfunc

def reduce(func,iterable,initializer=None):
    args = iter(iterable)
    if initializer is not None:
        res = initializer
    else:
        res = next(args)
    while True:
        try:
            res = func(res,next(args))
        except StopIteration:
            return res



"""Access to Python's configuration information."""

#well emulate this module since it does with settings very close to the
#OS and metal

variables={'TANH_PRESERVES_ZERO_SIGN': 0, 'WITH_DOC_STRINGS': 0}

def get_config_var(var):
    if var in variables:
       return variables[var]

    raise NotImplementedError("sysconfig.py:get_config_var: variable '%s' does not exist" % variable)



"""This module provides access to operating system functionality that is
standardized by the C Standard and the POSIX standard (a thinly
disguised Unix interface).  Refer to the library manual and
corresponding Unix manual entries for more information on calls."""

import datetime

from browser import window

def _randint(a, b):
    return int(window.Math.random()*(b-a+1)+a)
    
F_OK = 0

O_APPEND = 8

O_BINARY = 32768

O_CREAT = 256

O_EXCL = 1024

O_NOINHERIT = 128

O_RANDOM = 16

O_RDONLY = 0

O_RDWR = 2

O_SEQUENTIAL = 32

O_SHORT_LIVED = 4096

O_TEMPORARY = 64

O_TEXT = 16384

O_TRUNC = 512

O_WRONLY = 1

P_DETACH = 4

P_NOWAIT = 1

P_NOWAITO = 3

P_OVERLAY = 2

P_WAIT = 0

R_OK = 4

TMP_MAX = 32767

W_OK = 2

X_OK = 1

class __loader__:
    pass

def _exit(*args,**kw):
    """_exit(status)    
    Exit to the system with specified status, without normal exit processing."""
    pass

def _getdiskusage(*args,**kw):
    """_getdiskusage(path) -> (total, free)    
    Return disk usage statistics about the given path as (total, free) tuple."""
    pass

def _getfileinformation(*args,**kw):
    pass

def _getfinalpathname(*args,**kw):
    pass

def _getfullpathname(*args,**kw):
    pass

_have_functions = ['MS_WINDOWS']

def _isdir(*args,**kw):
    """Return true if the pathname refers to an existing directory."""
    pass

def abort(*args,**kw):
    """abort() -> does not return!    
    Abort the interpreter immediately.  This 'dumps core' or otherwise fails
    in the hardest way possible on the hosting operating system."""
    pass

def access(*args,**kw):
    """access(path, mode, *, dir_fd=None, effective_ids=False, follow_symlinks=True)    
    Use the real uid/gid to test for access to a path.  Returns True if granted,
    False otherwise.
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    If effective_ids is True, access will use the effective uid/gid instead of
      the real uid/gid.
    If follow_symlinks is False, and the last element of the path is a symbolic
      link, access will examine the symbolic link itself instead of the file the
      link points to.
    dir_fd, effective_ids, and follow_symlinks may not be implemented
      on your platform.  If they are unavailable, using them will raise a
      NotImplementedError.
    
    Note that most operations will use the effective uid/gid, therefore this
      routine can be used in a suid/sgid environment to test if the invoking user
      has the specified access to the path.
    The mode argument can be F_OK to test existence, or the inclusive-OR
      of R_OK, W_OK, and X_OK."""
    pass

def chdir(*args,**kw):
    """chdir(path)    
    Change the current working directory to the specified path.
    
    path may always be specified as a string.
    On some platforms, path may also be specified as an open file descriptor.
      If this functionality is unavailable, using it raises an exception."""
    pass

def chmod(*args,**kw):
    """chmod(path, mode, *, dir_fd=None, follow_symlinks=True)    
    Change the access permissions of a file.
    
    path may always be specified as a string.
    On some platforms, path may also be specified as an open file descriptor.
      If this functionality is unavailable, using it raises an exception.
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    If follow_symlinks is False, and the last element of the path is a symbolic
      link, chmod will modify the symbolic link itself instead of the file the
      link points to.
    It is an error to use dir_fd or follow_symlinks when specifying path as
      an open file descriptor.
    dir_fd and follow_symlinks may not be implemented on your platform.
      If they are unavailable, using them will raise a NotImplementedError."""
    pass

def close(*args,**kw):
    """close(fd)    
    Close a file descriptor (for low level IO)."""
    pass

def closerange(*args,**kw):
    """closerange(fd_low, fd_high)    
    Closes all file descriptors in [fd_low, fd_high), ignoring errors."""
    pass

def device_encoding(*args,**kw):
    """device_encoding(fd) -> str    
    Return a string describing the encoding of the device
    if the output is a terminal; else return None."""
    pass

def dup(*args,**kw):
    """dup(fd) -> fd2    
    Return a duplicate of a file descriptor."""
    pass

def dup2(*args,**kw):
    """dup2(old_fd, new_fd)    
    Duplicate file descriptor."""
    pass

environ = {'PYTHONUSERBASE': ' '}

error = OSError

def execv(*args,**kw):
    """execv(path, args)    
    Execute an executable path with arguments, replacing current process.
    
        path: path of executable file
        args: tuple or list of strings"""
    pass

def execve(*args,**kw):
    """execve(path, args, env)    
    Execute a path with arguments and environment, replacing current process.
    
        path: path of executable file
        args: tuple or list of arguments
        env: dictionary of strings mapping to strings
    
    On some platforms, you may specify an open file descriptor for path;
      execve will execute the program the file descriptor is open to.
      If this functionality is unavailable, using it raises NotImplementedError."""
    pass

def fstat(*args,**kw):
    """fstat(fd) -> stat result    
    Like stat(), but for an open file descriptor.
    Equivalent to stat(fd=fd)."""
    pass

def fsync(*args,**kw):
    """fsync(fildes)    
    force write of file with filedescriptor to disk."""
    pass

def get_terminal_size(*args,**kw):
    """Return the size of the terminal window as (columns, lines).    
    The optional argument fd (default standard output) specifies
    which file descriptor should be queried.
    
    If the file descriptor is not connected to a terminal, an OSError
    is thrown.
    
    This function will only be defined if an implementation is
    available for this system.
    
    shutil.get_terminal_size is the high-level function which should 
    normally be used, os.get_terminal_size is the low-level implementation."""
    pass

def getcwd(*args,**kw):
    """getcwd() -> path    
    Return a unicode string representing the current working directory."""
    return __BRYTHON__.brython_path # XXX fix me

def getcwdb(*args,**kw):
    """getcwdb() -> path    
    Return a bytes string representing the current working directory."""
    pass

def getlogin(*args,**kw):
    """getlogin() -> string    
    Return the actual login name."""
    pass

def getpid(*args,**kw):
    """getpid() -> pid    
    Return the current process id"""
    return 0

def getppid(*args,**kw):
    """getppid() -> ppid    
    Return the parent's process id.  If the parent process has already exited,
    Windows machines will still return its id; others systems will return the id
    of the 'init' process (1)."""
    pass

def isatty(*args,**kw):
    """isatty(fd) -> bool    
    Return True if the file descriptor 'fd' is an open file descriptor
    connected to the slave end of a terminal."""
    pass

def kill(*args,**kw):
    """kill(pid, sig)    
    Kill a process with a signal."""
    pass

def link(*args,**kw):
    """link(src, dst, *, src_dir_fd=None, dst_dir_fd=None, follow_symlinks=True)    
    Create a hard link to a file.
    
    If either src_dir_fd or dst_dir_fd is not None, it should be a file
      descriptor open to a directory, and the respective path string (src or dst)
      should be relative; the path will then be relative to that directory.
    If follow_symlinks is False, and the last element of src is a symbolic
      link, link will create a link to the symbolic link itself instead of the
      file the link points to.
    src_dir_fd, dst_dir_fd, and follow_symlinks may not be implemented on your
      platform.  If they are unavailable, using them will raise a
      NotImplementedError."""
    pass

def listdir(*args,**kw):
    """listdir(path='.') -> list_of_filenames    
    Return a list containing the names of the files in the directory.
    The list is in arbitrary order.  It does not include the special
    entries '.' and '..' even if they are present in the directory.
    
    path can be specified as either str or bytes.  If path is bytes,
      the filenames returned will also be bytes; in all other circumstances
      the filenames returned will be str.
    On some platforms, path may also be specified as an open file descriptor;
      the file descriptor must refer to a directory.
      If this functionality is unavailable, using it raises NotImplementedError."""
    pass

def lseek(*args,**kw):
    """lseek(fd, pos, how) -> newpos    
    Set the current position of a file descriptor.
    Return the new cursor position in bytes, starting from the beginning."""
    pass

def lstat(*args,**kw):
    """lstat(path, *, dir_fd=None) -> stat result    
    Like stat(), but do not follow symbolic links.
    Equivalent to stat(path, follow_symlinks=False)."""
    return stat_result()

def mkdir(*args,**kw):
    """mkdir(path, mode=0o777, *, dir_fd=None)    
    Create a directory.
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    dir_fd may not be implemented on your platform.
      If it is unavailable, using it will raise a NotImplementedError.
    
    The mode argument is ignored on Windows."""
    pass

def open(path, flags, mode=0o777, *args, dir_fd=None):
    """open(path, flags, mode=0o777, *, dir_fd=None)    
    Open a file for low level IO.  Returns a file handle (integer).
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    dir_fd may not be implemented on your platform.
      If it is unavailable, using it will raise a NotImplementedError."""
    
    ## lets assume this is reading/writing to a local storage in the browser
    from browser.local_storage import storage
    
    class mystorage:
      def __init__(self, path, flags):
          self._path=path
          self._pos=0
          self._flags=flags

          if self._flags & O_RDONLY == O_RDONLY:
             self._data=storage.get(self._path, None)
             if self._data is None:
                raise FileNotFoundError("%s not found" % self._path)
          elif self._flags & O_WRONLY == O_WRONLY:
             storage[self._path]=''

      def seek(self, pos):
          self._pos=pos

      def read(self, size=None):
          if size is None:
             _result=self._data[self._pos:]
             self._pos=len(self._data)
             return _result

          assert size <= len(self._data) - self._pos
          _result=self._data[self._pos: self._pos+size]
          self._pos+=size
          return _result

      def write(self, data):
          storage[self._path]+=str(data)

      def close(self):
          pass

    return mystorage(path, flags)

def pipe(*args,**kw):
    """pipe() -> (read_end, write_end)    
    Create a pipe."""
    pass

def putenv(*args,**kw):
    """putenv(key, value)    
    Change or add an environment variable."""
    pass

def read(*args,**kw):
    """read(fd, buffersize) -> string    
    Read a file descriptor."""
    pass

def readlink(*args,**kw):
    """readlink(path, *, dir_fd=None) -> path    
    Return a string representing the path to which the symbolic link points.
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    dir_fd may not be implemented on your platform.
      If it is unavailable, using it will raise a NotImplementedError."""
    pass

def remove(*args,**kw):
    """remove(path, *, dir_fd=None)    
    Remove a file (same as unlink()).
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    dir_fd may not be implemented on your platform.
      If it is unavailable, using it will raise a NotImplementedError."""
    pass

def rename(*args,**kw):
    """rename(src, dst, *, src_dir_fd=None, dst_dir_fd=None)    
    Rename a file or directory.
    
    If either src_dir_fd or dst_dir_fd is not None, it should be a file
      descriptor open to a directory, and the respective path string (src or dst)
      should be relative; the path will then be relative to that directory.
    src_dir_fd and dst_dir_fd, may not be implemented on your platform.
      If they are unavailable, using them will raise a NotImplementedError."""
    pass

def replace(*args,**kw):
    """replace(src, dst, *, src_dir_fd=None, dst_dir_fd=None)    
    Rename a file or directory, overwriting the destination.
    
    If either src_dir_fd or dst_dir_fd is not None, it should be a file
      descriptor open to a directory, and the respective path string (src or dst)
      should be relative; the path will then be relative to that directory.
    src_dir_fd and dst_dir_fd, may not be implemented on your platform.
      If they are unavailable, using them will raise a NotImplementedError."""
    pass

def rmdir(*args,**kw):
    """rmdir(path, *, dir_fd=None)    
    Remove a directory.
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    dir_fd may not be implemented on your platform.
      If it is unavailable, using it will raise a NotImplementedError."""
    pass

def spawnv(*args,**kw):
    """spawnv(mode, path, args)    
    Execute the program 'path' in a new process.
    
        mode: mode of process creation
        path: path of executable file
        args: tuple or list of strings"""
    pass

def spawnve(*args,**kw):
    """spawnve(mode, path, args, env)    
    Execute the program 'path' in a new process.
    
        mode: mode of process creation
        path: path of executable file
        args: tuple or list of arguments
        env: dictionary of strings mapping to strings"""
    pass

def startfile(*args,**kw):
    """startfile(filepath [, operation]) - Start a file with its associated    application.
    
    When "operation" is not specified or "open", this acts like
    double-clicking the file in Explorer, or giving the file name as an
    argument to the DOS "start" command: the file is opened with whatever
    application (if any) its extension is associated.
    When another "operation" is given, it specifies what should be done with
    the file.  A typical operation is "print".
    
    startfile returns as soon as the associated application is launched.
    There is no option to wait for the application to close, and no way
    to retrieve the application's exit status.
    
    The filepath is relative to the current directory.  If you want to use
    an absolute path, make sure the first character is not a slash ("/");
    the underlying Win32 ShellExecute function doesn't work if it is."""
    pass

def stat(*args,**kw):
    """stat(path, *, dir_fd=None, follow_symlinks=True) -> stat result    
    Perform a stat system call on the given path.
    
    path may be specified as either a string or as an open file descriptor.
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
      dir_fd may not be supported on your platform; if it is unavailable, using
      it will raise a NotImplementedError.
    If follow_symlinks is False, and the last element of the path is a symbolic
      link, stat will examine the symbolic link itself instead of the file the
      link points to.
    It is an error to use dir_fd or follow_symlinks when specifying path as
      an open file descriptor."""
    return stat_result()

def stat_float_times(*args,**kw):
    """stat_float_times([newval]) -> oldval    
    Determine whether os.[lf]stat represents time stamps as float objects.
    If newval is True, future calls to stat() return floats, if it is False,
    future calls return ints. 
    If newval is omitted, return the current setting.
    """
    pass

class stat_result:

    def __init__(self):
        """st_mode - protection bits, 
        st_ino - inode number, 
        st_dev - device, 
        st_nlink - number of hard links, 
        st_uid - user id of owner, 
        st_gid - group id of owner, 
        st_size - size of file, in bytes, 
        st_atime - time of most recent access expressed in seconds, 
        st_mtime - time of most recent content modification expressed in 
            seconds, 
        st_ctime - platform dependent; time of most recent metadata change on 
            Unix, or the time of creation on Windows, expressed in seconds 
        st_atime_ns - time of most recent access expressed in nanoseconds as an
             integer, 
        st_mtime_ns - time of most recent content modification expressed in 
            nanoseconds as an integer, 
        st_ctime_ns - platform dependent; time of most recent metadata change 
            on Unix, or the time of creation on Windows, expressed in 
            nanoseconds as an integer """
        # Brython : fake values
        self.st_atime = datetime.datetime.now()
        self.st_mtime = self.st_ctime = self.st_atime_ns = \
            self.st_mtime_ns = self.st_ctime_ns = self.st_atime
        self.st_uid = self.st_gid = self.st_ino = -1
        self.st_mode = 0
        self.st_size = 1

class statvfs_result:
    pass

def strerror(*args,**kw):
    """strerror(code) -> string    
    Translate an error code to a message string."""
    pass

def symlink(*args,**kw):
    """symlink(src, dst, target_is_directory=False, *, dir_fd=None)    
    Create a symbolic link pointing to src named dst.
    
    target_is_directory is required on Windows if the target is to be
      interpreted as a directory.  (On Windows, symlink requires
      Windows 6.0 or greater, and raises a NotImplementedError otherwise.)
      target_is_directory is ignored on non-Windows platforms.
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    dir_fd may not be implemented on your platform.
      If it is unavailable, using it will raise a NotImplementedError."""
    pass

def system(*args,**kw):
    """system(command) -> exit_status    
    Execute the command (a string) in a subshell."""
    pass

class terminal_size:
    pass

def times(*args,**kw):
    """times() -> times_result    
    Return an object containing floating point numbers indicating process
    times.  The object behaves like a named tuple with these fields:
      (utime, stime, cutime, cstime, elapsed_time)"""
    pass

class times_result:
    pass

def umask(*args,**kw):
    """umask(new_mask) -> old_mask    
    Set the current numeric umask and return the previous umask."""
    pass

class uname_result:
    pass

def unlink(path, *args, dir_fd=None):
    """unlink(path, *, dir_fd=None)    
    Remove a file (same as remove()).
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    dir_fd may not be implemented on your platform.
      If it is unavailable, using it will raise a NotImplementedError."""
    pass

def urandom(n):
    """urandom(n) -> str    
    Return n random bytes suitable for cryptographic use."""
    randbytes= [_randint(0,255) for i in range(n)]
    return bytes(randbytes)

def utime(*args,**kw):
    """utime(path, times=None, *, ns=None, dir_fd=None, follow_symlinks=True)    Set the access and modified time of path.
    
    path may always be specified as a string.
    On some platforms, path may also be specified as an open file descriptor.
      If this functionality is unavailable, using it raises an exception.
    
    If times is not None, it must be a tuple (atime, mtime);
        atime and mtime should be expressed as float seconds since the epoch.
    If ns is not None, it must be a tuple (atime_ns, mtime_ns);
        atime_ns and mtime_ns should be expressed as integer nanoseconds
        since the epoch.
    If both times and ns are None, utime uses the current time.
    Specifying tuples for both times and ns is an error.
    
    If dir_fd is not None, it should be a file descriptor open to a directory,
      and path should be relative; path will then be relative to that directory.
    If follow_symlinks is False, and the last element of the path is a symbolic
      link, utime will modify the symbolic link itself instead of the file the
      link points to.
    It is an error to use dir_fd or follow_symlinks when specifying path
      as an open file descriptor.
    dir_fd and follow_symlinks may not be available on your platform.
      If they are unavailable, using them will raise a NotImplementedError."""
    pass

def waitpid(*args,**kw):
    """waitpid(pid, options) -> (pid, status << 8)    
    Wait for completion of a given process.  options is ignored on Windows."""
    pass

def write(*args,**kw):
    """write(fd, string) -> byteswritten    
    Write a string to a file descriptor."""
    pass

## put WIFSIGNALED here. its needed by os module, and os module imports all
## functions in this module
def WIFSIGNALED(a):
    return False

def WTERMSIG(status):
    return 0

def WIFSIGNALED(status):
    "Return True if the process exited due to a signal, otherwise return False"
    return False

def WIFEXITED(status):
    return False

def WEXITSTATUS(status):
    pass

def WNOHANG():
    return (0,0)



from javascript import JSObject
from browser import window
import urllib.request

class TempMod:
  def __init__(self, name):
      self.name=name

#define my custom import hook (just to see if it get called etc).
class BaseHook:
  def __init__(self, fullname=None, path=None):
      self._fullname=fullname
      self._path=path    # we don't are about this...
      self._modpath=''
      self._module=''

  def find_module(self, name=None, path=None):
      if name is None:
         name=self._fullname

      for _i in ('libs/%s.js' % name, 'Lib/%s.py' % name, 
                 'Lib/%s/__init__.py' % name):
          _path="%s%s" % (__BRYTHON__.brython_path, _i)
          try:
            _fp,_,_headers=urllib.request.urlopen(_path)
            if _headers['status'] != 200:
               continue 
            self._module=_fp.read()
            self._modpath=_path
            return self
          except urllib.error.HTTPError as e:
            self._modpath=''
            self._module=''
            
      raise ImportError

  def is_package(self):
      return '.' in self._fullname

  def load_module(self, name):
      if name is None:
         name=self._fullname
      window.eval('__BRYTHON__.imported["%s"]={}' % name)
      return JSObject(__BRYTHON__.run_py)(self._module,
                                          self._modpath, TempMod(name))



import logging

from browser.ajax import ajax


class XMLHTTPHandler(logging.Handler):
    """
    A class which sends records to a Web server, using either GET or
    POST semantics.
    """
    def __init__(self, url, method="GET"):
        """
        Initialize the instance with the host, the request URL, and the method
        ("GET" or "POST")
        """
        logging.Handler.__init__(self)
        method = method.upper()
        if method not in ["GET", "POST"]:
            raise ValueError("method must be GET or POST")
        self.url = url
        self.method = method

    def mapLogRecord(self, record):
        """
        Default implementation of mapping the log record into a dict
        that is sent as the CGI data. Overwrite in your class.
        Contributed by Franz Glasner.
        """
        return record.__dict__

    def emit(self, record):
        """
        Emit a record.

        Send the record to the Web server as a percent-encoded dictionary
        """
        try:
            req = ajax.open(self.method, self.url, async=False)
            req.send(self.mapLogRecord(record))
        except:
            self.handleError(record)



#
# Module providing the `Process` class which emulates `threading.Thread`
#
# multiprocessing/process.py
#
# Copyright (c) 2006-2008, R Oudkerk
# Licensed to PSF under a Contributor Agreement.
#

__all__ = ['Process', 'current_process', 'active_children']

#
# Imports
#

import os
import sys
import signal
import itertools
from _weakrefset import WeakSet

#for brython
from _multiprocessing import Process
#
#
#

try:
    ORIGINAL_DIR = os.path.abspath(os.getcwd())
except OSError:
    ORIGINAL_DIR = None

#
# Public functions
#

def current_process():
    '''
    Return process object representing the current process
    '''
    return _current_process

def active_children():
    '''
    Return list of process objects corresponding to live child processes
    '''
    _cleanup()
    return list(_current_process._children)

#
#
#

def _cleanup():
    # check for processes which have finished
    for p in list(_current_process._children):
        if p._popen.poll() is not None:
            _current_process._children.discard(p)

#
# The `Process` class
#

# brython note: class Process is defined in /usr/libs/_multiprocessing.js


#
# We subclass bytes to avoid accidental transmission of auth keys over network
#

class AuthenticationString(bytes):
    def __reduce__(self):
        from .forking import Popen
        if not Popen.thread_is_spawning():
            raise TypeError(
                'Pickling an AuthenticationString object is '
                'disallowed for security reasons'
                )
        return AuthenticationString, (bytes(self),)

#
# Create object representing the main process
#

class _MainProcess(Process):

    def __init__(self):
        self._identity = ()
        self._daemonic = False
        self._name = 'MainProcess'
        self._parent_pid = None
        self._popen = None
        self._counter = itertools.count(1)
        self._children = set()
        self._authkey = AuthenticationString(os.urandom(32))
        self._tempdir = None

_current_process = _MainProcess()
del _MainProcess

#
# Give names to some return codes
#

_exitcode_to_name = {}

for name, signum in list(signal.__dict__.items()):
    if name[:3]=='SIG' and '_' not in name:
        _exitcode_to_name[-signum] = name

# For debug and leak testing
_dangling = WeakSet()



"""allow programmer to define multiple exit functions to be executedupon normal program termination.

Two public functions, register and unregister, are defined.
"""


class __loader__(object):
    pass

def _clear(*args,**kw):
    """_clear() -> None    
    Clear the list of previously registered exit functions."""
    pass

def _run_exitfuncs(*args,**kw):
    """_run_exitfuncs() -> None    
    Run all registered exit functions."""
    pass

def register(*args,**kw):
    """register(func, *args, **kwargs) -> func    
    Register a function to be executed upon normal program termination
    
        func - function to be called at exit
        args - optional arguments to pass to func
        kwargs - optional keyword arguments to pass to func
    
        func is returned to facilitate usage as a decorator."""
    pass

def unregister(*args,**kw):
    """unregister(func) -> None    
    Unregister a exit function which was previously registered using
    atexit.register
    
        func - function to be unregistered"""
    pass



"""This module provides mechanisms to use signal handlers in Python.

Functions:

alarm() -- cause SIGALRM after a specified time [Unix only]
setitimer() -- cause a signal (described below) after a specified
               float time and the timer may restart then [Unix only]
getitimer() -- get current value of timer [Unix only]
signal() -- set the action for a given signal
getsignal() -- get the signal action for a given signal
pause() -- wait until a signal arrives [Unix only]
default_int_handler() -- default SIGINT handler

signal constants:
SIG_DFL -- used to refer to the system default handler
SIG_IGN -- used to ignore the signal
NSIG -- number of defined signals
SIGINT, SIGTERM, etc. -- signal numbers

itimer constants:
ITIMER_REAL -- decrements in real time, and delivers SIGALRM upon
               expiration
ITIMER_VIRTUAL -- decrements only when the process is executing,
               and delivers SIGVTALRM upon expiration
ITIMER_PROF -- decrements both when the process is executing and
               when the system is executing on behalf of the process.
               Coupled with ITIMER_VIRTUAL, this timer is usually
               used to profile the time spent by the application
               in user and kernel space. SIGPROF is delivered upon
               expiration.


*** IMPORTANT NOTICE ***
A signal handler function is called with two arguments:
the first is the signal number, the second is the interrupted stack frame."""

CTRL_BREAK_EVENT=1
CTRL_C_EVENT=0
NSIG=23
SIGABRT=22
SIGBREAK=21
SIGFPE=8
SIGILL=4
SIGINT=2
SIGSEGV=11
SIGTERM=15
SIG_DFL=0
SIG_IGN=1

def signal(signalnum, handler) :
    pass



__author__ = 'Loris Mularoni'


# Import modules
import os
import logging
import configparser


# Path of the R script to create the plot
REGULOME_R_SCRIPT = os.path.join(os.path.dirname(__file__), 'regulome_plot', 'plot_IRB_main.R')


# Read the configuration file. It creates is if it does not exist.
configs = configparser.ConfigParser()
configs._interpolation = configparser.ExtendedInterpolation()
_ = configs.read(os.path.join(os.getcwd(), 'regulome.cfg'))


class Config:
    SECRET_KEY = configs['secret_key']['key']
    UPLOAD_FOLDER = configs['data']['uploads']
    # Maximum dimension of the uploaded file. If a larger file is transmitted,
    # Flask will raise an RequestEntityTooLarge exception.
    MAX_CONTENT_LENGTH = 50 * 1024 * 1024  # 50Mb.
    # SQLALCHEMY_DATABASE_URI = configs['databases']['genes']
    # SQLALCHEMY_BINDS = {
    #     'regions': configs['databases']['regions'],
    #     'tfbs': configs['databases']['tfbs']
    # }
    # SQLALCHEMY_TRACK_MODIFICATIONS = True


class Development(Config):
    DEBUG = True


class Testing(Config):
    TESTING = True


class Production(Config):
    DEBUG = False


web_configuration = {
    'development': Development,
    'testing': Testing,
    'production': Production,
    'default': Development
}


# Logger
# Set the verbosity of the logger. Levels are:
# critical 50
# error	 40
# warning	 30
# info	 20
# debug	 10
# notset	 0

levels = {
    'critical': logging.CRITICAL,
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'info': logging.INFO,
    'debug': logging.DEBUG,
    'notset': logging.NOTSET
}

# Set the logger
logger = logging.getLogger(__name__)
logger.setLevel(levels.get(configs['logs']['log_level'], logging.INFO))
formatter = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s', datefmt='%H:%M')

# File handler
fh = logging.FileHandler(configs['logs']['regulome_log'])
fh.setLevel(levels.get(configs['logs']['log_level'], logging.INFO))
fh.setFormatter(formatter)
logger.addHandler(fh)

# Stdout handler
ch = logging.StreamHandler()
ch.setLevel(levels.get(configs['logs']['log_level'], logging.INFO))
ch.setFormatter(formatter)
logger.addHandler(ch)
